from __future__ import annotations

import attrs
import mitsuba as mi
import numpy as np
import pint

from ._distant import AbstractDistantMeasure
from ... import frame, validators
from ...attrs import documented, parse_docs
from ...config import settings
from ...units import symbol
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...warp import square_to_uniform_hemisphere


@parse_docs
@attrs.define(eq=False, slots=False)
class DistantFluxMeasure(AbstractDistantMeasure):
    """
    Distant radiosity measure scene element [``distantflux``, ``distant_flux``].

    This measure records the [spectral] radiosity (in W/m²[/nm]) at an infinite
    distance in a hemisphere defined by its ``direction`` parameter.

    When used with a backward tracing algorithm, rays traced by the sensor
    target a shape which can be controlled through the ``target`` parameter.
    This feature is useful if one wants to compute the average flux leaving
    a particular subset of the scene.

    Notes
    -----
    * Setting the ``target`` parameter is required to get meaningful results.
      Experiment classes should take care of setting it appropriately.
    * The film resolution can be adjusted to manually stratify film sampling
      and reduce variance in results. The default 32x32 is generally a good
      choice, but scenes with sharp reflection lobes may benefit from higher
      values.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    azimuth_convention: frame.AzimuthConvention = documented(
        attrs.field(
            default=None,
            converter=lambda x: (
                settings.azimuth_convention
                if x is None
                else (frame.AzimuthConvention[x.upper()] if isinstance(x, str) else x)
            ),
            validator=attrs.validators.instance_of(frame.AzimuthConvention),
        ),
        doc="Azimuth convention. If ``None``, the global default configuration "
        "is used (see :ref:`sec-user_guide-config`).",
        type=".AzimuthConvention",
        init_type=".AzimuthConvention or str, optional",
        default="None",
    )

    direction: np.ndarray = documented(
        attrs.field(
            default=[0, 0, 1],
            converter=np.array,
            validator=validators.is_vector3,
        ),
        doc="A 3-vector defining the normal to the reference surface for which "
        "the exitant flux density is computed.",
        type="array-like",
        default="[0, 0, 1]",
    )

    _film_resolution: tuple[int, int] = documented(
        attrs.field(
            default=(32, 32),
            validator=attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(int),
                iterable_validator=validators.has_len(2),
            ),
        ),
        doc="Film resolution as a (width, height) 2-tuple.",
        type="array-like",
        default="(32, 32)",
    )

    @property
    def film_resolution(self) -> tuple[int, int]:
        return self._film_resolution

    @property
    def viewing_angles(self) -> pint.Quantity:
        """
        quantity: Viewing angles computed from stored film coordinates as a
            (width, height, 2) array. The last dimension is ordered as
            (zenith, azimuth).
        """
        # Compute viewing angles at pixel locations
        # Angle computation must match the kernel plugin's direction sampling
        # routine
        angle_units = ucc.get("angle")

        # Compute pixel locations in film coordinates
        xs = (
            np.linspace(0, 1, self.film_resolution[0], endpoint=False)
            + 0.5 / self.film_resolution[0]
        )
        ys = (
            np.linspace(0, 1, self.film_resolution[1], endpoint=False)
            + 0.5 / self.film_resolution[1]
        )

        # Compute corresponding angles in specified azimuth convention
        xy = np.array([(x, y) for x in xs for y in ys])
        angles = frame.direction_to_angles(
            square_to_uniform_hemisphere(xy),
            azimuth_convention=self.azimuth_convention,
        ).to(angle_units)

        # Reshape array to match film size on first 2 dimensions
        return angles.reshape((len(xs), len(ys), 2))

    # --------------------------------------------------------------------------
    #                        Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def kernel_type(self) -> str:
        # Inherit docstring
        return "distantflux"

    @property
    def template(self) -> dict:
        # Inherit docstring
        result = super().template

        _, up = mi.coordinate_system(self.direction)
        result["to_world"] = mi.ScalarTransform4f.look_at(
            origin=[0, 0, 0],
            target=mi.ScalarVector3f(self.direction),
            up=up,
        )

        if self.target is not None:
            result["target"] = self.target.kernel_item()

        if self.ray_offset is not None:
            result["ray_offset"] = self.ray_offset.m_as(uck.get("length"))

        return result

    # --------------------------------------------------------------------------
    #                        Post-processing information
    # --------------------------------------------------------------------------

    @property
    def var(self) -> tuple[str, dict]:
        # Inherit docstring
        return "sector_radiosity", {
            "standard_name": "sector_radiosity",
            "long_name": "sector radiosity",
            "units": symbol(uck.get("irradiance")),
        }



from __future__ import annotations

import itertools
import typing as t
from collections.abc import MutableMapping
from copy import deepcopy

import attrs
import numpy as np
import pint
import pinttr

from ._core import Canopy, InstancedCanopyElement, biosphere_factory
from ._leaf_cloud import CuboidLeafCloudParams, LeafCloud
from ...attrs import documented, parse_docs
from ...units import unit_context_config as ucc


def _instanced_canopy_elements_converter(value):
    """
    Special converter for the DiscreteCanopy.instanced_canopy_elements field.
    """
    if isinstance(value, MutableMapping):
        value["type"] = "instanced"
    return biosphere_factory.convert(value)


@parse_docs
@attrs.define(eq=False, slots=False)
class DiscreteCanopy(Canopy):
    """
    Discrete canopy scene element [``discrete_canopy``].

    A discrete canopy composed of one or several :class:`.CanopyElement`
    instances. The underlying implementation holds a list of
    :class:`.InstancedCanopyElement`, allowing for arbitrary cloning of each
    :class:`.CanopyElement`. The produced canopy can be padded with clones of
    itself using the :meth:`~.DiscreteCanopy.padded_copy` method.

    .. admonition:: Class method constructors

       .. autosummary::

          homogeneous
          leaf_cloud_from_files
          padded
    """

    # --------------------------------------------------------------------------
    #                                   Fields
    # --------------------------------------------------------------------------

    instanced_canopy_elements: list[InstancedCanopyElement] = documented(
        attrs.field(
            factory=list,
            converter=lambda value: [
                _instanced_canopy_elements_converter(x)
                for x in pinttr.util.always_iterable(value)
            ]
            if not isinstance(value, MutableMapping)
            else [_instanced_canopy_elements_converter(value)],
            validator=attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(InstancedCanopyElement)
            ),
        ),
        doc="List of :class:`.CanopyElement` defining the canopy. Can be "
        "initialised with a :class:`.InstancedCanopyElement`, which will be "
        "automatically wrapped into a list. Dictionary-based specifications are "
        "allowed as well.",
        type="list of :class:`.InstancedCanopyElement`",
        default="[]",
    )

    # --------------------------------------------------------------------------
    #                          Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def _template_bsdfs(self) -> dict:
        result = {}
        for element in self.instanced_canopy_elements:
            result.update(element._template_bsdfs)
        return result

    @property
    def _template_shapes(self) -> dict:
        result = {}
        for element in self.instanced_canopy_elements:
            result.update(element._template_shapes)
        return result

    @property
    def _template_instances(self) -> dict:
        result = {}
        for element in self.instanced_canopy_elements:
            result.update(element._template_instances)
        return result

    @property
    def template(self) -> dict:
        result = {}

        for element in self.instanced_canopy_elements:
            result.update(
                {
                    **element._template_bsdfs,
                    **element._template_shapes,
                    **element._template_instances,
                }
            )

        return result

    @property
    def _params_instances(self) -> dict:
        result = {}
        for element in self.instanced_canopy_elements:
            result.update(element._params_instances)
        return result

    @property
    def params(self) -> dict:
        result = {}

        for element in self.instanced_canopy_elements:
            result.update(
                {
                    **element._params_bsdfs,
                    **element._params_shapes,
                    **element._params_instances,
                }
            )

        return result

    # --------------------------------------------------------------------------
    #                                  Padding
    # --------------------------------------------------------------------------

    def padded_copy(self, padding: int) -> DiscreteCanopy:
        """
        Return a copy of the current canopy padded with additional copies.

        Parameters
        ----------
        padding : int
            Amount of padding around the canopy. Must be positive or zero.
            The resulting padded canopy is a grid of
            :math:`2 \\times \\mathit{padding} + 1` copies.

        Returns
        -------
        .DiscreteCanopy
            Padded copy of self.
        """
        if padding < 0:
            raise ValueError("padding must be >= 0")

        if padding == 0:
            return self

        # We'll return a copy
        result = deepcopy(self)

        # Convenience aliases
        config_length = ucc.get("length")
        x_size, y_size = result.size.m_as(config_length)[:2]
        padding_factors = np.array(list(range(-padding, padding + 1)))

        for instanced_canopy_element in result.instanced_canopy_elements:
            # More convenience aliases
            old_instance_positions = instanced_canopy_element.instance_positions.m_as(
                config_length
            )
            n_instances_per_cell = old_instance_positions.shape[0]
            # Allocate array for results
            new_instance_positions = np.empty(
                (len(padding_factors) ** 2 * n_instances_per_cell, 3)
            )

            for k, (x_offset_factor, y_offset_factor) in enumerate(
                itertools.product(padding_factors, padding_factors)
            ):
                # Set vector by which we will translate instance positions
                offset = np.array(
                    [x_size * x_offset_factor, y_size * y_offset_factor, 0.0]
                )

                # Compute new instances
                start_idx = k * n_instances_per_cell
                stop_idx = (k + 1) * n_instances_per_cell
                new_instance_positions[start_idx:stop_idx, :] = (
                    old_instance_positions[:, :] + offset
                )

            instanced_canopy_element.instance_positions = (
                new_instance_positions * config_length
            )

        # Update size
        result.size[:2] *= len(padding_factors)

        return result

    # --------------------------------------------------------------------------
    #                               Constructors
    # --------------------------------------------------------------------------

    @classmethod
    def padded(cls, padding: int = 0, **kwargs) -> DiscreteCanopy:
        """
        Create a discrete canopy and pad it with copies of itself. Keyword
        arguments are forwarded to the default constructor.

        Parameters
        ----------
        padding : int
            Amount of padding around the canopy. Must be positive or zero.
            The resulting padded canopy is a grid of
            :math:`2 \\times \\mathit{padding} + 1` copies.

        **kwargs``
            Keyword arguments forwarded to the :class:`.DiscreteCanopy`
            constructor.

        Returns
        -------
        :class:`.DiscreteCanopy`
            Padded discrete canopy.
        """
        return cls(**kwargs).padded_copy(padding)

    @classmethod
    def homogeneous(
        cls,
        padding: int = 0,
        id: str = "homogeneous_discrete_canopy",
        **leaf_cloud_kwargs,
    ) -> DiscreteCanopy:
        """
        Generate a homogeneous discrete canopy, possibly padded with copies of
        itself.

        Parameters
        ----------
        padding : int
            Amount of padding around the canopy. Must be positive or zero.
            The resulting padded canopy is a grid of
            :math:`2 \\times \\mathit{padding} + 1` copies.

        id : str, default: "homogeneous_discrete_canopy"
            Canopy object ID.

        **leaf_cloud_kwargs
            Keyword arguments forwarded to :meth:`.LeafCloud.cuboid`.

        Returns
        -------
        .DiscreteCanopy
            Created canopy object.

        Notes
        -----
        The leaf cloud's ID will be set to ``f"{id}_leaf_cloud"``.
        """
        # Check parameters
        leaf_cloud_params = CuboidLeafCloudParams(**leaf_cloud_kwargs)
        config_length = ucc.get("length")
        size = [
            leaf_cloud_params.l_horizontal.m_as(config_length),
            leaf_cloud_params.l_horizontal.m_as(config_length),
            leaf_cloud_params.l_vertical.m_as(config_length),
        ] * config_length
        leaf_cloud_id = f"{id}_leaf_cloud"

        # Construct canopy
        return cls(
            id=id,
            size=size,
            instanced_canopy_elements=[
                InstancedCanopyElement(
                    instance_positions=[[0, 0, 0]],
                    canopy_element=LeafCloud.cuboid(
                        **leaf_cloud_kwargs, id=leaf_cloud_id
                    ),
                )
            ],
        ).padded_copy(padding)

    @classmethod
    def leaf_cloud_from_files(
        cls,
        size: pint.Quantity,
        leaf_cloud_dicts: list[t.MutableMapping],
        padding: int = 0,
        id: str = "discrete_canopy",
    ):
        """
        Directly create a leaf cloud canopy from text file specifications,
        possibly padded with copies of itself.

        .. admonition:: CanopyElement dictionary format

           Each item of the ``leaf_cloud_dicts`` list shall have the following
           structure:

           .. code:: python

              {
                  "sub_id": "some_value",  # leaf cloud ID string part, optional if leaf_cloud_dicts has only 1 entry
                  "instance_filename": "some_path",  # path to instance specification file
                  "leaf_cloud_filename": "some_other_path", # path to leaf cloud specification file
                  "leaf_reflectance": 0.5,  # optional, leaf reflectance (default: 0.5)
                  "leaf_transmittance": 0.5,  # optional, leaf transmittance (default: 0.5)
              }

        Parameters
        ----------
        size : quantity
            Canopy size as a 3-vector (in metres).

        leaf_cloud_dicts : list of dict
            List of dictionary specifying canopy elements and instances (see
            format above).

        padding : int, default: 0
            Amount of padding around the canopy. Must be positive or zero.
            The resulting padded canopy is a grid of
            :math:`2 \\times \\mathit{padding} + 1` copies.

        id : str, default: "discrete_canopy"
            Canopy ID.

        Returns
        -------
        .DiscreteCanopy
            Created canopy object.
        """
        instanced_canopy_elements = []

        for leaf_cloud_dict in leaf_cloud_dicts:
            instance_filename = leaf_cloud_dict.get("instance_filename", None)

            leaf_cloud_params = {
                "filename": leaf_cloud_dict.get("leaf_cloud_filename", None),
                "leaf_reflectance": leaf_cloud_dict.get("leaf_reflectance", 0.5),
                "leaf_transmittance": leaf_cloud_dict.get("leaf_transmittance", 0.5),
            }

            sub_id = leaf_cloud_dict.get("sub_id", None)

            if sub_id is None:
                if len(leaf_cloud_dicts) > 1:
                    raise ValueError("parameter 'sub_id' must be set")
                leaf_cloud_params["id"] = f"{id}_leaf_cloud"
            else:
                leaf_cloud_params["id"] = f"{id}_{sub_id}_leaf_cloud"

            instanced_canopy_elements.append(
                InstancedCanopyElement.from_file(
                    filename=instance_filename,
                    canopy_element=LeafCloud.from_file(**leaf_cloud_params),
                )
            )

        return cls(
            id=id,
            size=size,
            instanced_canopy_elements=instanced_canopy_elements,
        ).padded_copy(padding)



"""
Particle layers.
"""

from __future__ import annotations

from functools import singledispatchmethod

import attrs
import numpy as np
import pint
import pinttr
import xarray as xr

from ._core import AbstractHeterogeneousAtmosphere
from ._particle_dist import ParticleDistribution, particle_distribution_factory
from ..core import traverse
from ..phase import TabulatedPhaseFunction
from ... import converters, data
from ...attrs import documented, parse_docs
from ...contexts import KernelContext
from ...kernel import UpdateParameter
from ...radprops import ZGrid
from ...spectral.ckd import BinSet, QuadSpec
from ...spectral.index import (
    CKDSpectralIndex,
    MonoSpectralIndex,
    SpectralIndex,
)
from ...spectral.mono import WavelengthSet
from ...units import to_quantity
from ...units import unit_context_config as ucc
from ...units import unit_registry as ureg
from ...util.misc import cache_by_id, summary_repr
from ...validators import is_positive


def _particle_layer_distribution_converter(value):
    if isinstance(value, str):
        if value == "uniform":
            return particle_distribution_factory.convert({"type": "uniform"})
        elif value == "gaussian":
            return particle_distribution_factory.convert({"type": "gaussian"})
        elif value == "exponential":
            return particle_distribution_factory.convert({"type": "exponential"})

    return particle_distribution_factory.convert(value)


@parse_docs
@attrs.define(eq=False, slots=False)
class ParticleLayer(AbstractHeterogeneousAtmosphere):
    """
    Particle layer scene element [``particle_layer``].

    The particle layer has a vertical extension specified by a bottom altitude
    (set by ``bottom``) and a top altitude (set by ``top``).
    Inside the layer, the particles number is distributed according to a
    distribution (set by ``distribution``).
    See :mod:`~eradiate.scenes.atmosphere.particle_dist` for the available
    distribution types and corresponding parameters.
    The particle layer is itself divided into a number of (sub-)layers
    (``n_layers``) to allow to describe the variations of the particles number
    with altitude.
    The particle density in the layer is adjusted so that the particle layer's
    optical thickness at a specified reference wavelength (``w_ref``) meets a
    specified value (``tau_ref``).
    The particles radiative properties are specified by a data set
    (``dataset``).
    """

    bottom: pint.Quantity = documented(
        pinttr.field(
            default=ureg.Quantity(0.0, ureg.km),
            validator=[
                is_positive,
                pinttr.validators.has_compatible_units,
            ],
            units=ucc.deferred("length"),
        ),
        doc="Bottom altitude of the particle layer.\n"
        "\n"
        "Unit-enabled field (default: ucc[length])",
        type="quantity",
        init_type="float or quantity",
        default="0 km",
    )

    top: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("length"),
            default=ureg.Quantity(1.0, ureg.km),
            validator=[
                is_positive,
                pinttr.validators.has_compatible_units,
            ],
        ),
        doc="Top altitude of the particle layer.\n"
        "\n"
        "Unit-enabled field (default: ucc[length]).",
        type="quantity",
        init_type="float or quantity",
        default="1 km.",
    )

    @bottom.validator
    @top.validator
    def _bottom_top_validator(self, attribute, value):
        if self.bottom >= self.top:
            raise ValueError(
                f"while validating '{attribute.name}': bottom altitude must be "
                "lower than top altitude "
                f"(got bottom={self.bottom}, top={self.top})"
            )

    distribution: ParticleDistribution = documented(
        attrs.field(
            default="uniform",
            converter=_particle_layer_distribution_converter,
            validator=attrs.validators.instance_of(ParticleDistribution),
        ),
        doc="Particle distribution. Simple defaults can be set using a string: "
        '``"uniform"`` (resp. ``"gaussian"``, ``"exponential"``) is converted to '
        ":class:`UniformParticleDistribution() <.UniformParticleDistribution>` "
        "(resp. :class:`GaussianParticleDistribution() <.GaussianParticleDistribution>`, "
        ":class:`ExponentialParticleDistribution() <.ExponentialParticleDistribution>`).",
        init_type=":class:`.ParticleDistribution` or dict or "
        '{"uniform", "gaussian", "exponential"}, optional',
        type=":class:`.ParticleDistribution`",
        default='"uniform"',
    )

    w_ref: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("wavelength"),
            default=550 * ureg.nm,
            validator=[
                is_positive,
                pinttr.validators.has_compatible_units,
            ],
        ),
        doc="Reference wavelength at which the extinction optical thickness is "
        "specified.\n"
        "\n"
        "Unit-enabled field (default: ucc['wavelength']).",
        type="quantity",
        init_type="quantity or float",
        default="550.0",
    )

    tau_ref: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("dimensionless"),
            default=ureg.Quantity(0.2, ureg.dimensionless),
            validator=[
                is_positive,
                pinttr.validators.has_compatible_units,
            ],
        ),
        doc="Extinction optical thickness at the reference wavelength.\n"
        "\n"
        "Unit-enabled field (default: ucc[dimensionless]).",
        type="quantity",
        init_type="quantity or float",
        default="0.2",
    )

    dataset: xr.Dataset = documented(
        attrs.field(
            default="govaerts_2021-continental",
            converter=converters.to_dataset(
                load_from_id=lambda x: data.load_dataset(f"spectra/particles/{x}.nc")
            ),
            validator=attrs.validators.instance_of(xr.Dataset),
            repr=summary_repr,
        ),
        doc="Particle radiative property data set."
        "If an xarray dataset is passed, the dataset is used as is "
        "(refer to the data guide for the format requirements of this dataset)."
        "If a path is passed, the converter tries to open the corresponding "
        "file on the hard drive; should that fail, it queries the Eradiate data"
        "store with that path."
        "If a string is passed, it is interpreted as an identifier for a "
        "particle radiative property dataset in the Eradiate data store.",
        type="Dataset",
        init_type="Dataset or path-like or str, optional",
        default="govaerts_2021-continental",
    )

    has_absorption: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
        ),
        doc="Absorption bypass switch. If ``True``, the absorption coefficient "
        "is computed. Else, the absorption coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
        ),
        doc="Scattering bypass switch. If ``True``, the scattering coefficient "
        "is computed. Else, the scattering coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    @has_absorption.validator
    @has_scattering.validator
    def _switch_validator(self, attribute, value):
        if not self.has_absorption and not self.has_scattering:
            raise ValueError(
                f"while validating {attribute.name}: at least one of "
                "'has_absorption' and 'has_scattering' must be True"
            )

    _phase: TabulatedPhaseFunction | None = attrs.field(default=None, init=False)

    def spectral_set(
        self, quad_spec: QuadSpec | None = None
    ) -> None | BinSet | WavelengthSet:
        return None

    def update(self) -> None:
        self._phase = TabulatedPhaseFunction(id=self.phase_id, data=self.dataset.phase)

    # --------------------------------------------------------------------------
    #                    Spatial and thermophysical properties
    # --------------------------------------------------------------------------

    def eval_fractions(self, zgrid: ZGrid) -> np.ndarray:
        """
        Compute the particle number fraction in the particle layer.

        Returns
        -------
        ndarray
            Particle number fractions as a (n_layers,)-shaped array.
        """
        x = (zgrid.layers - self.bottom) / (self.top - self.bottom)
        fractions = self.distribution(x.m_as(ureg.dimensionless))
        fractions /= np.sum(fractions)
        return fractions

    def eval_mfp(self, ctx: KernelContext) -> pint.Quantity:
        min_sigma_s = self.eval_sigma_s(ctx.si).min()
        return 1.0 / min_sigma_s if min_sigma_s != 0.0 else np.inf * ureg.m

    # --------------------------------------------------------------------------
    #                       Radiative properties
    # --------------------------------------------------------------------------

    @cache_by_id
    def _eval_albedo_impl(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Return albedo from dataset (without accounting for bypass switches)
        # This routine is vectorized and returns an array of shape
        # (n_wavelengths, n_layers)
        ds = self.dataset
        wavelengths = w.m_as(ds.w.attrs["units"])
        interpolated = to_quantity(ds.albedo.interp(w=wavelengths))
        where_present = np.reshape(self.eval_fractions(zgrid) > 0, (1, -1))
        return interpolated * where_present

    @cache_by_id
    def _eval_sigma_t_impl(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Return extinction coefficient from dataset (without accounting
        # for bypass switches). This routine is vectorized and returns an
        # array of shape (n_wavelengths, n_layers)

        # Collect input data
        ds = self.dataset
        ds_w_units = ureg(ds.w.attrs["units"])
        wavelengths = np.atleast_1d(w.m_as(ds_w_units))
        sigma_t_star = to_quantity(ds.sigma_t.interp(w=wavelengths))
        sigma_t_star_ref = to_quantity(ds.sigma_t.interp(w=self.w_ref.m_as(ds_w_units)))

        # Compute target optical thickness value
        tau = self.tau_ref * sigma_t_star / sigma_t_star_ref

        # Scatter this total OT to all layers
        # TODO: Make sure that axis order is consistent with other vectorized
        #  routines
        fractions = self.eval_fractions(zgrid)
        tau_layers = np.broadcast_to(
            np.reshape(tau, (-1, 1)), (len(wavelengths), zgrid.n_layers)
        ) * np.reshape(fractions, (1, -1))

        # Compute corresponding average coefficient
        sigma_t = tau_layers / zgrid.layer_height

        return sigma_t

    def _eval_sigma_a_impl(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Return absorption coefficient from dataset (without accounting for
        # bypass switches). This routine is vectorized and returns an array of
        # shape (n_wavelengths, n_layers)
        albedo = self._eval_albedo_impl(w, zgrid)
        return self._eval_sigma_t_impl(w, zgrid) * (1.0 - albedo.m)

    def _eval_sigma_s_impl(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Return scattering coefficient from dataset (without accounting for
        # bypass switches). This routine is vectorized and returns an array of
        # shape (n_wavelengths, n_layers)
        albedo = self._eval_albedo_impl(w, zgrid)
        return self._eval_sigma_t_impl(w, zgrid) * albedo.m

    @singledispatchmethod
    def eval_albedo(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        raise NotImplementedError

    @eval_albedo.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_albedo_mono(
            w=si.w,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    @eval_albedo.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_albedo_ckd(
            w=si.w,
            g=si.g,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_albedo_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        if self.has_absorption and self.has_scattering:
            albedo = self._eval_albedo_impl(w, zgrid).squeeze()

        elif self.has_absorption and not self.has_scattering:
            albedo = 0.0 * ureg.dimensionless

        elif self.has_scattering and not self.has_absorption:
            albedo = 1.0 * ureg.dimensionless

        else:
            raise RuntimeError

        # Albedo is constant vs spatial dimension
        return np.full_like(zgrid.layers, albedo)

    def eval_albedo_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        return self.eval_albedo_mono(w=w, zgrid=zgrid)

    @singledispatchmethod
    def eval_sigma_t(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
    ) -> pint.Quantity:
        # Inherit docstring
        raise NotImplementedError

    @eval_sigma_t.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_t_mono(
            w=si.w,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    @eval_sigma_t.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_t_ckd(
            w=si.w,
            g=si.g,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_t_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        result = self._eval_sigma_t_impl(w, zgrid).squeeze()

        if self.has_absorption and self.has_scattering:
            return result

        elif not self.has_absorption and self.has_scattering:
            return result - self._eval_sigma_a_impl(w, zgrid)

        elif self.has_absorption and not self.has_scattering:
            return result - self._eval_sigma_s_impl(w, zgrid)

        raise RuntimeError

    def eval_sigma_t_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        return self.eval_sigma_t_mono(w=w, zgrid=zgrid)

    @singledispatchmethod
    def eval_sigma_a(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
    ) -> pint.Quantity:
        # Inherit docstring
        raise NotImplementedError

    @eval_sigma_a.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_a_mono(
            w=si.w,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    @eval_sigma_a.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_a_ckd(
            w=si.w,
            g=si.g,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_a_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        value = self._eval_sigma_a_impl(w, zgrid).squeeze()
        return value if self.has_absorption else np.zeros_like(value) * value.units

    def eval_sigma_a_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        return self.eval_sigma_a_mono(w, zgrid)

    @singledispatchmethod
    def eval_sigma_s(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        raise NotImplementedError

    @eval_sigma_s.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_s_mono(
            w=si.w,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    @eval_sigma_s.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid | None = None) -> pint.Quantity:
        return self.eval_sigma_s_ckd(
            w=si.w,
            g=si.g,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_s_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        value = self._eval_sigma_s_impl(w, zgrid).squeeze()
        return value if self.has_scattering else np.zeros_like(value) * value.units

    def eval_sigma_s_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        return self.eval_sigma_s_mono(w, zgrid)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def phase(self) -> TabulatedPhaseFunction:
        # Inherit docstring
        return self._phase

    @property
    def _template_phase(self) -> dict:
        result, _ = traverse(self.phase)
        return result.data

    @property
    def _params_phase(self) -> dict[str, UpdateParameter]:
        _, result = traverse(self.phase)
        return result.data



import xarray as xr

from ..units import symbol


def make_particle_dataset(albedo, sigma_t, phase, mu, w):
    return xr.Dataset(
        data_vars={
            "sigma_t": (
                "w",
                sigma_t.magnitude,
                dict(
                    standard_name="air_volume_extinction_coefficient",
                    units=symbol(sigma_t.units),
                ),
            ),
            "albedo": (
                "w",
                albedo.magnitude,
                dict(
                    standard_name="single_scattering_albedo",
                    units=symbol(albedo.units),
                ),
            ),
            "phase": (
                ("w", "mu", "i", "j"),
                phase.magnitude,
                dict(
                    standard_name="scattering_phase_matrix",
                    units=symbol(phase.units),
                ),
            ),
        },
        coords={
            "w": ("w", w.magnitude, dict(units=symbol(w.units))),
            "mu": (
                "mu",
                mu.magnitude,
                dict(
                    standard_name="scattering_angle_cosine",
                    units=symbol(mu.units),
                ),
            ),
            "i": ("i", [0]),
            "j": ("j", [0]),
        },
    )



import numpy as np

from eradiate import unit_registry as ureg
from eradiate.experiments import AtmosphereExperiment


def create_rami4atm_hom00_bla_sd2s_m03_z30a000_brfpp():
    r"""
    RAMI4ATM HOM00_BLA_SD2S_M03
    ===========================

    Rationale
    ---------
    This test case uses a basic atmospheric scene:

    * black surface
    * Molecular atmosphere following the AFGL 1986 model
    * Aerosol layer at 1km elevation

    Parameters

    * Atmosphere: molecular atmosphere using the AFGL 1986 profile
      (U.S. standard)
    * Aerosol layer: uniform, covers [0, 2] km extent, :math:`\tau_{500} = 0.2`;
      radiative properties from the ``govaerts_2021-desert`` dataset
    * Surface: black
    * Illumination: directional illumination with a zenith angle
      :math:`\theta = 30°`
    * Measure: distant measure, covering the principal plane for
      :math:`\theta \in [-75, 75]°`
    """

    config = {
        "surface": {
            "reflectance": {"value": 0.0, "type": "uniform"},
            "type": "lambertian",
        },
        "atmosphere": {
            "molecular_atmosphere": {
                "has_absorption": False,
                "has_scattering": True,
                "type": "molecular",
                "thermoprops": {
                    "identifier": "afgl_1986-us_standard",
                    "z": np.arange(0, 120.05, 0.05) * ureg.km,
                },
                "absorption_data": "monotropa",
            },
            "particle_layers": [
                {
                    "bottom": 0,
                    "bottom_units": "meter",
                    "top": 2000,
                    "top_units": "meter",
                    "distribution": {"type": "uniform"},
                    "tau_ref": 0.2,
                    "dataset": "govaerts_2021-desert",
                }
            ],
            "type": "heterogeneous",
        },
        "illumination": {
            "zenith": 30.0,
            "zenith_units": "degree",
            "azimuth": 0.0,
            "azimuth_units": "degree",
            "type": "directional",
        },
        "measures": [
            {
                "type": "mdistant",
                "construct": "hplane",
                "zeniths": np.arange(-75, 76, 2),
                "zeniths_units": "degree",
                "azimuth": 0.0,
                "azimuth_units": "degree",
                "srf": "sentinel_2a-msi-3",
                "spp": 1000,
            }
        ],
        "quad_spec": {
            "type": "fixed",
            "quad_type": "gauss_legendre",
            "n": 16,
        },
    }

    return AtmosphereExperiment(**config)
    



from __future__ import annotations

import logging
import typing as t
import warnings

import attrs
import mitsuba as mi
from tqdm.auto import tqdm

from ._kernel_dict import UpdateMapTemplate
from .. import config
from ..attrs import documented, parse_docs
from ..contexts import KernelContext
from ..rng import SeedState, root_seed_state

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
#                         Parameter lookup strategies
# ------------------------------------------------------------------------------


@parse_docs
@attrs.frozen
class TypeIdLookupStrategy:
    """
    This parameter ID lookup strategy searches for a Mitsuba type and object ID
    match.

    Instances are callables which take, as argument, the current node during
    a Mitsuba scene tree traversal and, optionally, its path in the Mitsuba
    scene tree. If the lookup succeeds, the full parameter path is returned.
    """

    node_type: type = documented(
        attrs.field(validator=attrs.validators.instance_of(type)),
        doc="Type of the node which will be looked up.",
        type="type",
    )

    node_id: str = documented(
        attrs.field(validator=attrs.validators.instance_of(str)),
        doc="ID of the node which will be looked up.",
        type="str",
    )

    parameter_relpath: str = documented(
        attrs.field(validator=attrs.validators.instance_of(str)),
        doc="Parameter path relative to its parent object.",
        type="str",
    )

    def __call__(self, node, node_path: str | None = None) -> str | None:
        if isinstance(node, self.node_type) and node.id() == self.node_id:
            prefix = f"{node_path}." if node_path is not None else ""
            return f"{prefix}{self.parameter_relpath}"

        else:
            return None


# ------------------------------------------------------------------------------
#                           Mitsuba scene traversal
# ------------------------------------------------------------------------------


@parse_docs
@attrs.define
class MitsubaObjectWrapper:
    """
    This container aggregates a Mitsuba object, its associated parameters and a
    set of updaters that can be used to modify the scene parameters.

    See Also
    --------
    :func:`mi_traverse`
    """

    obj: mi.Object = documented(
        attrs.field(repr=lambda x: "Scene[...]" if isinstance(x, mi.Scene) else str(x)),
        doc="Mitsuba object.",
        type="mitsuba.Object",
    )

    parameters: mi.SceneParameters | None = documented(
        attrs.field(
            default=None,
            repr=lambda x: (
                "SceneParameters[...]"
                if x.__class__.__name__ == "SceneParameters"
                else str(None)
            ),
        ),
        doc="Mitsuba scene parameter map.",
        type="mitsuba.SceneParameters",
        init_type="mitsuba.SceneParameters, optional",
        default="None",
    )

    umap_template: UpdateMapTemplate | None = documented(
        attrs.field(
            default=None,
            repr=lambda x: (
                "UpdateMapTemplate[...]" if isinstance(x, UpdateMapTemplate) else str(x)
            ),
        ),
        doc="An update map template, which can be rendered and used to update "
        "Mitsuba scene parameters depending on context information.",
        type=".UpdateMapTemplate",
        init_type=".UpdateMapTemplate, optional",
        default="None",
    )

    def drop_parameters(self) -> None:
        """
        Reduce the size of the scene parameter table :attr:`.parameters` by
        only keeping elements whose keys are listed in the parameter update
        map template :attr:`.umap_template`. For parameters associated with a
        lookup protocol, the looked up parameter ID is checked and used.
        """
        if self.umap_template is not None:
            keys = []
            for name, param in self.umap_template.items():
                if param.lookup_strategy is not None:
                    if param.parameter_id is not None:
                        keys.append(param.parameter_id)
                    else:
                        warnings.warn(
                            f"Parameter '{name}' has a lookup strategy but the "
                            "associated parameter ID is undefined; was a "
                            "parameter lookup performed during the Mitsuba "
                            "scene traversal?"
                        )
                else:
                    keys.append(name)

            self.parameters.keep(keys)


def mi_traverse(
    obj: mi.Object,
    umap_template: UpdateMapTemplate | None = None,
) -> MitsubaObjectWrapper:
    """
    Traverse a node of the Mitsuba scene graph and return a container holding
    the Mitsuba scene, its parameter map and an updated parameter update map.

    Parameters
    ----------
    obj : mitsuba.Object
        Mitsuba scene graph node to be traversed.

    umap_template : .UpdateMapTemplate, optional
        An additional update map template which is to be updated during
        traversal. This is used to perform parameter lookup during traversal.

    Returns
    -------
    MitsubaObjectWrapper
        A container holding the traversed object, the corresponding parameter
        map and the parameter update map (if any).

    Notes
    -----
    This is a reimplementation of the :func:`mitsuba.traverse` function.
    """

    umap_template = (
        UpdateMapTemplate(data=umap_template.data.copy())
        if umap_template is not None
        else UpdateMapTemplate()
    )

    lookups = {
        k: v
        for k, v in umap_template.items()
        if v.parameter_id is None and v.lookup_strategy is not None
    }

    class SceneTraversal(mi.TraversalCallback):
        def __init__(
            self,
            node,
            parent=None,
            properties=None,
            hierarchy=None,
            prefixes=None,
            name=None,
            depth=0,
            flags=+mi.ParamFlags.Differentiable,
        ):
            mi.TraversalCallback.__init__(self)
            self.properties = dict() if properties is None else properties
            self.hierarchy = dict() if hierarchy is None else hierarchy
            self.prefixes = set() if prefixes is None else prefixes

            if name is not None:
                ctr, name_len = 1, len(name)
                while name in self.prefixes:
                    name = "%s_%i" % (name[:name_len], ctr)
                    ctr += 1
                self.prefixes.add(name)

            self.name = name
            self.node = node
            self.depth = depth
            self.hierarchy[node] = (parent, depth)
            self.flags = flags

            # Try and recover a parameter ID from this node
            for name, uparam in list(lookups.items()):
                lookup_result = uparam.lookup_strategy(self.node, self.name)
                if lookup_result is not None:
                    uparam.parameter_id = lookup_result
                    del lookups[
                        name
                    ]  # Remove successful lookups to accelerate future searches

        def put_parameter(self, name, ptr, flags, cpptype=None):
            name = name if self.name is None else self.name + "." + name

            flags = self.flags | flags
            # Non-differentiable parameters shouldn't be flagged as discontinuous
            if (flags & mi.ParamFlags.NonDifferentiable) != 0:
                flags = flags & ~mi.ParamFlags.Discontinuous

            self.properties[name] = (ptr, cpptype, self.node, self.flags | flags)

        def put_object(self, name, node, flags):
            if node is None or node in self.hierarchy:
                return
            cb = SceneTraversal(
                node=node,
                parent=self.node,
                properties=self.properties,
                hierarchy=self.hierarchy,
                prefixes=self.prefixes,
                name=name if self.name is None else f"{self.name}.{name}",
                depth=self.depth + 1,
                flags=self.flags | flags,
            )
            node.traverse(cb)

    cb = SceneTraversal(obj)
    obj.traverse(cb)

    # Check if there are unsuccessful lookups
    if lookups:
        warnings.warn(
            "There were unsuccessful Mitsuba scene parameter lookups: "
            f"{list(lookups.keys())}"
        )

    return MitsubaObjectWrapper(
        obj=obj,
        parameters=mi.SceneParameters(cb.properties, cb.hierarchy),
        umap_template=umap_template,
    )


# ------------------------------------------------------------------------------
#                             Mitsuba scene render
# ------------------------------------------------------------------------------


def mi_render(
    mi_scene: MitsubaObjectWrapper,
    ctxs: list[KernelContext],
    sensors: None | int | list[int] = None,
    spp: int = 0,
    seed_state: SeedState | None = None,
) -> dict[t.Any, mi.Bitmap]:
    """
    Render a Mitsuba scene multiple times given specified contexts and sensor
    indices.

    Parameters
    ----------
    mi_scene : .MitsubaObjectWrapper
        Mitsuba scene to render.

    ctxs : list of :class:`.KernelContext`
        List of contexts used to generate the parameter update table at each
        iteration.

    sensors : int or list of int, optional
        Sensor indices to render. If ``None`` (default), all sensors are
        rendered.

    spp : int, optional, default: 0
        Number of samples per pixel. If set to 0 (default), the value set in the
        original scene definition takes precedence.

    seed_state : .SeedState, optional
        Seed state used to generate seeds to initialize Mitsuba's RNG at
        each run. If unset, Eradiate's root seed state is used.

    Returns
    -------
    dict
        A nested dictionary mapping context and sensor indices to rendered
        bitmaps.

    Notes
    -----
    This function wraps sequential calls to  :func:`mitsuba.render`.
    """

    if seed_state is None:
        logger.debug("Using default RNG seed generator")
        seed_state = root_seed_state

    results = {}

    # Loop on contexts
    with tqdm(
        initial=0,
        total=len(ctxs),
        unit_scale=1.0,
        leave=True,
        bar_format="{desc}{n:g}/{total:g}|{bar}| {elapsed}, ETA={remaining}",
        disable=(config.settings.progress < config.ProgressLevel.SPECTRAL_LOOP)
        or len(ctxs) <= 1,
    ) as pbar:
        for ctx in ctxs:
            pbar.set_description(
                f"Eradiate [{ctx.index_formatted}]",
                refresh=True,
            )

            logger.debug("Updating Mitsuba scene parameters")
            mi_scene.parameters.update(mi_scene.umap_template.render(ctx))

            if sensors is None:
                mi_sensors = [
                    (i, sensor) for i, sensor in enumerate(mi_scene.obj.sensors())
                ]

            else:
                if isinstance(sensors, int):
                    sensors = [sensors]
                mi_sensors = [(i, mi_scene.obj.sensors()[i]) for i in sensors]

            # Loop on sensors
            for i_sensor, mi_sensor in mi_sensors:
                # Render sensor
                seed = int(seed_state.next().squeeze())
                logger.debug(
                    'Running Mitsuba for sensor "%s" with seed value %s',
                    mi_sensor.id(),
                    seed,
                )
                mi.render(mi_scene.obj, sensor=i_sensor, seed=seed, spp=spp)

                # Store result in a new Bitmap object
                siah = ctx.si.as_hashable
                if siah not in results:
                    results[siah] = {}

                results[siah][mi_sensor.id()] = mi.Bitmap(mi_sensor.film().bitmap())

            pbar.update()

    return results



from __future__ import annotations

import typing as t
from abc import ABC
from copy import deepcopy

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttr
from pinttr.util import ensure_units

from ._core import Measure
from ... import converters, frame, validators
from ...attrs import documented, parse_docs
from ...config import settings
from ...units import symbol
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg
from ...util.misc import is_vector3

# ------------------------------------------------------------------------------
#                             Measure target interface
# ------------------------------------------------------------------------------


@attrs.define
class Target:
    """
    Interface for target selection objects used by distant measure classes.
    """

    def kernel_item(self) -> dict:
        """Return kernel item."""
        raise NotImplementedError

    @staticmethod
    def new(target_type, *args, **kwargs) -> Target:
        """
        Instantiate one of the supported child classes. This factory requires
        manual class registration. All position and keyword arguments are
        forwarded to the constructed type.

        Currently supported classes:

        * ``point``: :class:`.TargetPoint`
        * ``rectangle``: :class:`.TargetRectangle`

        Parameters
        ----------
        target_type : {"point", "rectangle"}
            Identifier of one of the supported child classes.

        Returns
        -------
        :class:`.Target`
        """
        if target_type == "point":
            return TargetPoint(*args, **kwargs)
        elif target_type == "rectangle":
            return TargetRectangle(*args, **kwargs)
        else:
            raise ValueError(f"unknown target type {target_type}")

    @staticmethod
    def convert(value) -> t.Any:
        """
        Object converter method.

        If ``value`` is a dictionary, this method uses :meth:`new` to
        instantiate a :class:`Target` child class based on the ``"type"`` entry
        it contains.

        If ``value`` is a 3-vector, this method returns a :class:`.TargetPoint`
        instance.

        Otherwise, it returns ``value``.
        """
        if isinstance(value, dict):
            d = deepcopy(value)
            try:
                target_type = d.pop("type")
            except KeyError:
                raise ValueError("cannot convert dict, missing 'type' entry")

            return Target.new(target_type, **d)

        if is_vector3(value):
            return Target.new("point", xyz=value)

        return value


def _target_point_rectangle_xyz_converter(x):
    return converters.on_quantity(float)(
        pinttr.converters.to_units(ucc.deferred("length"))(x)
    )


@parse_docs
@attrs.define
class TargetPoint(Target):
    """
    Point target or origin specification.
    """

    # Target point in config units
    xyz: pint.Quantity = documented(
        pinttr.field(units=ucc.deferred("length")),
        doc="Point coordinates.\n\nUnit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
    )

    @xyz.validator
    def _xyz_validator(self, attribute, value):
        if not is_vector3(value):
            raise ValueError(
                f"while validating {attribute.name}: must be a "
                f"3-element vector of numbers"
            )

    def kernel_item(self) -> dict:
        """Return kernel item."""
        return self.xyz.m_as(uck.get("length"))


@parse_docs
@attrs.define
class TargetRectangle(Target):
    """
    Rectangle target origin specification.

    This class defines an axis-aligned rectangular zone where ray targets will
    be sampled or ray origins will be projected.
    """

    xmin: pint.Quantity = documented(
        pinttr.field(
            converter=_target_point_rectangle_xyz_converter,
            units=ucc.deferred("length"),
        ),
        doc="Lower bound on the X axis.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="quantity or float",
    )

    xmax: pint.Quantity = documented(
        pinttr.field(
            converter=_target_point_rectangle_xyz_converter,
            units=ucc.deferred("length"),
        ),
        doc="Upper bound on the X axis.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="quantity or float",
    )

    ymin: pint.Quantity = documented(
        pinttr.field(
            converter=_target_point_rectangle_xyz_converter,
            units=ucc.deferred("length"),
        ),
        doc="Lower bound on the Y axis.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="quantity or float",
    )

    ymax: pint.Quantity = documented(
        pinttr.field(
            converter=_target_point_rectangle_xyz_converter,
            units=ucc.deferred("length"),
        ),
        doc="Upper bound on the Y axis.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="quantity or float",
    )

    z: pint.Quantity = documented(
        pinttr.field(
            default=0.0,
            converter=_target_point_rectangle_xyz_converter,
            units=ucc.deferred("length"),
        ),
        doc="Altitude of the plane enclosing the rectangle.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="quantity or float",
        default="0.0",
    )

    @xmin.validator
    @xmax.validator
    @ymin.validator
    @ymax.validator
    @z.validator
    def _xyz_validator(self, attribute, value):
        validators.on_quantity(validators.is_number)(self, attribute, value)

    @xmin.validator
    @xmax.validator
    def _x_validator(self, attribute, value):
        if not self.xmin < self.xmax:
            raise ValueError(
                f"while validating {attribute.name}: 'xmin' must "
                f"be lower than 'xmax"
            )

    @ymin.validator
    @ymax.validator
    def _y_validator(self, attribute, value):
        if not self.ymin < self.ymax:
            raise ValueError(
                f"while validating {attribute.name}: 'ymin' must "
                f"be lower than 'ymax"
            )

    def kernel_item(self) -> dict:
        # Inherit docstring

        kernel_length = uck.get("length")
        xmin = self.xmin.m_as(kernel_length)
        xmax = self.xmax.m_as(kernel_length)
        ymin = self.ymin.m_as(kernel_length)
        ymax = self.ymax.m_as(kernel_length)
        z = self.z.m_as(kernel_length)

        dx = xmax - xmin
        dy = ymax - ymin

        to_world = mi.ScalarTransform4f.translate(
            [0.5 * dx + xmin, 0.5 * dy + ymin, z]
        ) @ mi.ScalarTransform4f.scale([0.5 * dx, 0.5 * dy, 1.0])

        return {"type": "rectangle", "to_world": to_world}


# ------------------------------------------------------------------------------
#                             Distant measure interface
# ------------------------------------------------------------------------------


@parse_docs
@attrs.define(eq=False, slots=False)
class AbstractDistantMeasure(Measure, ABC):
    """
    Abstract interface of all distant measure classes.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    target: Target | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(Target.convert),
            validator=attrs.validators.optional(attrs.validators.instance_of(Target)),
            on_setattr=attrs.setters.pipe(
                attrs.setters.convert, attrs.setters.validate
            ),
        ),
        doc="Target specification. The target can be specified using an "
        "array-like with 3 elements (which will be converted to a "
        ":class:`.TargetPoint`) or a dictionary interpreted by "
        ":meth:`Target.convert() <.Target.convert>`. If set to "
        "``None`` (not recommended), the default target point selection "
        "method is used: rays will not target a particular region of the "
        "scene.",
        type=":class:`.Target` or None",
        init_type=":class:`.Target` or dict or array-like, optional",
    )

    ray_offset: pint.Quantity | None = documented(
        pinttr.field(default=None, units=ucc.deferred("length")),
        doc="Manually control the distance between the target and ray origins. "
        "If unset, ray origins are positioned outside of the scene and this "
        "measure is rigorously distant.",
        type="quantity or None",
        init_type="float or quantity, optional",
        default="None",
    )

    @ray_offset.validator
    def _ray_offset_validator(self, attribute, value):
        if value is None:
            return

        if value.magnitude <= 0:
            raise ValueError(
                f"while validating '{attribute.name}': only positive values "
                f"are allowed, got {value}"
            )

    # --------------------------------------------------------------------------
    #                             Flag-style queries
    # --------------------------------------------------------------------------

    def is_distant(self) -> bool:
        # Inherit docstring
        return self.ray_offset is None


@parse_docs
@attrs.define(eq=False, slots=False)
class DistantMeasure(AbstractDistantMeasure):
    """
    Single-pixel distant measure scene element [``distant``]

    This scene element records radiance leaving the scene in a single direction
    defined by its ``direction`` parameter. Most users will however find the
    :class:`.MultiDistantMeasure` class more flexible.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    azimuth_convention: frame.AzimuthConvention = documented(
        attrs.field(
            default=None,
            converter=lambda x: (
                settings.azimuth_convention
                if x is None
                else frame.AzimuthConvention.convert(x)
            ),
            validator=attrs.validators.instance_of(frame.AzimuthConvention),
        ),
        doc="Azimuth convention. If ``None``, the global default configuration "
        "is used (see :ref:`sec-user_guide-config`).",
        type=".AzimuthConvention",
        init_type=".AzimuthConvention or str, optional",
        default="None",
    )

    direction: np.ndarray = documented(
        attrs.field(
            default=[0, 0, 1],
            converter=np.array,
            validator=validators.is_vector3,
        ),
        doc="A 3-vector defining the direction observed by the sensor, pointing "
        "outwards the target.",
        type="ndarray",
        init_type="array-like",
        default="[0, 0, 1]",
    )

    @property
    def film_resolution(self) -> tuple[int, int]:
        return 1, 1

    @property
    def viewing_angles(self) -> pint.Quantity:
        """
        quantity: Viewing angles computed from the `direction` parameter as
            (1, 1, 2) array. The last dimension is ordered as (zenith, azimuth).
        """
        angles = frame.direction_to_angles(
            self.direction,
            azimuth_convention=self.azimuth_convention,
            normalize=True,
        ).to(ucc.get("angle"))  # Convert to default angle units
        return np.reshape(angles, (1, 1, 2))

    # --------------------------------------------------------------------------
    #                         Additional constructors
    # --------------------------------------------------------------------------

    @classmethod
    def from_angles(cls, angles: pint.Quantity, **kwargs) -> DistantMeasure:
        """
        Construct using a direction layout defined by explicit (zenith, azimuth)
        pairs.

        Parameters
        ----------
        angles : array-like
            A (zenith, azimuth) pair, either as a quantity or a unitless
            array-like. In the latter case, the default angle units are applied.

        azimuth_convention : .AzimuthConvention or str, optional
            The azimuth convention applying to the viewing direction layout.
            If unset, the global default convention is used.

        **kwargs
            Remaining keyword arguments are forwarded to the
            :class:`.DistantMeasure` constructor.

        Returns
        -------
        DistantMeasure
        """
        azimuth_convention = kwargs.pop("azimuth_convention", None)
        if azimuth_convention is None:
            azimuth_convention = settings.azimuth_convention

        angles = ensure_units(angles, default_units=ucc.get("angle")).m_as(ureg.rad)
        direction = np.squeeze(
            frame.angles_to_direction(
                angles=angles, azimuth_convention=azimuth_convention
            )
        )
        return cls(direction=direction, **kwargs)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def kernel_type(self) -> str:
        # Inherit docstring
        return "distant"

    @property
    def template(self) -> dict:
        # Inherit docstring
        result = super().template
        result["direction"] = mi.ScalarVector3f(-self.direction)

        if self.target is not None:
            result["target"] = self.target.kernel_item()

        if self.ray_offset is not None:
            result["ray_offset"] = self.ray_offset.m_as(uck.get("length"))

        return result

    @property
    def var(self) -> tuple[str, dict]:
        # Inherit docstring
        return "radiance", {
            "standard_name": "radiance",
            "long_name": "radiance",
            "units": symbol(uck.get("radiance")),
        }


@parse_docs
@attrs.define(eq=False, slots=False)
class MultiPixelDistantMeasure(AbstractDistantMeasure):
    """
    Multi-pixel distant measure scene element [``mpdistant``, ``multipixel_distant``]

    This scene element records radiance leaving the scene in a single direction
    defined by its ``direction`` parameter. Most users will however find the
    :class:`.MultiDistantMeasure` class more flexible.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    azimuth_convention: frame.AzimuthConvention = documented(
        attrs.field(
            default=None,
            converter=lambda x: (
                settings.azimuth_convention
                if x is None
                else frame.AzimuthConvention.convert(x)
            ),
            validator=attrs.validators.instance_of(frame.AzimuthConvention),
        ),
        doc="Azimuth convention. If ``None``, the global default configuration "
        "is used (see :ref:`sec-user_guide-config`).",
        type=".AzimuthConvention",
        init_type=".AzimuthConvention or str, optional",
        default="None",
    )

    direction: np.ndarray = documented(
        attrs.field(
            default=[0, 0, 1],
            converter=np.array,
            validator=validators.is_vector3,
        ),
        doc="A 3-vector defining the direction observed by the sensor, pointing "
        "outwards the target.",
        type="ndarray",
        init_type="array-like",
        default="[0, 0, 1]",
    )

    _film_resolution: tuple[int, int] = documented(
        attrs.field(
            default=(32, 32),
            validator=attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(int),
                iterable_validator=validators.has_len(2),
            ),
        ),
        doc="Film resolution as a (width, height) 2-tuple.",
        type="array-like",
        default="(32, 32)",
    )

    @property
    def film_resolution(self) -> tuple[int, int]:
        return self._film_resolution

    @property
    def viewing_angles(self) -> pint.Quantity:
        """
        quantity: Viewing angles computed from stored film coordinates as a
            (width, height, 2) array. The last dimension is ordered as
            (zenith, azimuth).
        """
        angles: pint.Quantity = frame.direction_to_angles(
            self.direction, azimuth_convention=self.azimuth_convention
        ).squeeze()
        shape = (*self.film_resolution, 2)
        return np.broadcast_to(angles.m, shape) * angles.u

    # --------------------------------------------------------------------------
    #                         Additional constructors
    # --------------------------------------------------------------------------

    @classmethod
    def from_angles(cls, angles: pint.Quantity, **kwargs) -> MultiPixelDistantMeasure:
        """
        Construct using a direction layout defined by explicit (zenith, azimuth)
        pairs.

        Parameters
        ----------
        angles : array-like
            A (zenith, azimuth) pair, either as a quantity or a unitless
            array-like. In the latter case, the default angle units are applied.

        azimuth_convention : .AzimuthConvention or str, optional
            The azimuth convention applying to the viewing direction layout.
            If unset, the global default convention is used.

        **kwargs
            Remaining keyword arguments are forwarded to the
            :class:`.DistantMeasure` constructor.

        Returns
        -------
        DistantMeasure
        """
        azimuth_convention = kwargs.pop("azimuth_convention", None)
        if azimuth_convention is None:
            azimuth_convention = settings.azimuth_convention

        angles = ensure_units(angles, default_units=ucc.get("angle")).m_as(ureg.rad)
        direction = np.squeeze(
            frame.angles_to_direction(
                angles=angles, azimuth_convention=azimuth_convention
            )
        )
        return cls(direction=direction, **kwargs)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def kernel_type(self) -> str:
        # Inherit docstring
        return "mpdistant"

    @property
    def template(self) -> dict:
        # Inherit docstring
        result = super().template
        result["direction"] = mi.ScalarVector3f(-self.direction)

        if self.target is not None:
            result["target"] = self.target.kernel_item()

        if self.ray_offset is not None:
            result["ray_offset"] = self.ray_offset.m_as(uck.get("length"))

        return result

    @property
    def var(self) -> tuple[str, dict]:
        # Inherit docstring
        return "radiance", {
            "standard_name": "radiance",
            "long_name": "radiance",
            "units": symbol(uck.get("radiance")),
        }



from __future__ import annotations

import typing as t
from pathlib import Path

import attrs

from ._core import DataStore
from ..attrs import documented, parse_docs
from ..exceptions import DataError
from ..typing import PathLike


@parse_docs
@attrs.define
class BlindDirectoryDataStore(DataStore):
    """
    Serve files stored in a directory.
    """

    path: Path = documented(
        attrs.field(converter=lambda x: Path(x).absolute()),
        type="Path",
        init_type="path-like",
        doc="Path to the root of the directory referenced by this data store.",
    )

    @property
    def base_url(self) -> str:
        """
        Raises :class:`NotImplementedError` (this data store has no target
        location).
        """
        raise NotImplementedError

    @property
    def registry(self) -> dict:
        """
        Raises :class:`NotImplementedError` (this data store has no registry).
        """
        raise NotImplementedError

    def registry_files(
        self, filter: t.Callable[[t.Any], bool] | None = None
    ) -> list[str]:
        """
        Returns an empty list (this data store has no registry).
        """
        return []

    def fetch(self, filename: PathLike, **kwargs) -> Path:
        # No kwargs are actually accepted
        if kwargs:
            keyword = next(iter(kwargs.keys()))
            raise TypeError(f"fetch() got an unexpected keyword argument '{keyword}'")

        fname = self.path / filename
        if fname.is_file():
            return fname
        else:
            raise DataError(f"file '{str(filename)}' is not in '{str(self.path)}'")



from ._core import Measure as Measure
from ._core import measure_factory as measure_factory
from ._distant import AbstractDistantMeasure as AbstractDistantMeasure
from ._distant import DistantMeasure as DistantMeasure
from ._distant import MultiPixelDistantMeasure as MultiPixelDistantMeasure
from ._distant import Target as Target
from ._distant import TargetPoint as TargetPoint
from ._distant import TargetRectangle as TargetRectangle
from ._distant_flux import DistantFluxMeasure as DistantFluxMeasure
from ._hemispherical_distant import (
    HemisphericalDistantMeasure as HemisphericalDistantMeasure,
)
from ._multi_distant import AngleLayout as AngleLayout
from ._multi_distant import AzimuthRingLayout as AzimuthRingLayout
from ._multi_distant import DirectionLayout as DirectionLayout
from ._multi_distant import GridLayout as GridLayout
from ._multi_distant import HemispherePlaneLayout as HemispherePlaneLayout
from ._multi_distant import Layout as Layout
from ._multi_distant import MultiDistantMeasure as MultiDistantMeasure
from ._multi_radiancemeter import MultiRadiancemeterMeasure as MultiRadiancemeterMeasure
from ._perspective import PerspectiveCameraMeasure as PerspectiveCameraMeasure
from ._radiancemeter import RadiancemeterMeasure as RadiancemeterMeasure



"""
Heterogeneous atmospheres.
"""
from __future__ import annotations

import logging
from collections import abc as cabc

import attrs
import mitsuba as mi
import numpy as np
import pint

from ._core import AbstractHeterogeneousAtmosphere, atmosphere_factory
from ._molecular import MolecularAtmosphere
from ._particle_layer import ParticleLayer
from ..core import traverse
from ..phase import BlendPhaseFunction, PhaseFunction
from ...attrs import documented, parse_docs
from ...contexts import KernelContext
from ...kernel import TypeIdLookupStrategy
from ...radprops import ZGrid
from ...spectral.ckd import BinSet, QuadSpec
from ...spectral.index import SpectralIndex
from ...spectral.mono import WavelengthSet
from ...units import unit_context_config as ucc
from ...units import unit_registry as ureg
from ...util.misc import cache_by_id

logger = logging.getLogger(__name__)


def _molecular_converter(value):
    if isinstance(value, cabc.MutableMapping) and ("type" not in value):
        value["type"] = "molecular"
    return atmosphere_factory.convert(value, allowed_cls=MolecularAtmosphere)


def _particle_layer_converter(value):
    if not value:
        return []

    if not isinstance(value, (list, tuple)):
        return _particle_layer_converter([value])

    else:
        result = []

        for element in value:
            if isinstance(element, cabc.MutableMapping) and ("type" not in element):
                element["type"] = "particle_layer"
            result.append(
                atmosphere_factory.convert(element, allowed_cls=ParticleLayer)
            )

        return result


@parse_docs
@attrs.define(eq=False, slots=False)
class HeterogeneousAtmosphere(AbstractHeterogeneousAtmosphere):
    """
    Heterogeneous atmosphere scene element [``heterogeneous``].
    """

    molecular_atmosphere: MolecularAtmosphere | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(_molecular_converter),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(MolecularAtmosphere)
            ),
        ),
        doc="Molecular atmosphere. May be specified as a dictionary interpreted "
        'by :data:`.atmosphere_factory`; in that case, the ``"type"`` parameter '
        'may be omitted and will automatically be set to ``"molecular"``.',
        type=".MolecularAtmosphere or None",
        init_type=".MolecularAtmosphere or dict, optional",
        default="None",
    )

    @molecular_atmosphere.validator
    def _molecular_atmosphere_validator(self, attribute, value):
        if value is None:
            return

        if value.scale is not None:
            raise ValueError(
                f"while validating {attribute.name}: components cannot be "
                "scaled individually"
            )

    particle_layers: list[ParticleLayer] = documented(
        attrs.field(
            factory=list,
            converter=_particle_layer_converter,
            validator=attrs.validators.deep_iterable(
                attrs.validators.instance_of(ParticleLayer)
            ),
        ),
        doc="List of particle layers. Elements may be specified as "
        "dictionaries interpreted by :data:`.atmosphere_factory`; in that "
        "case, the ``type`` parameter may be omitted and will automatically "
        'be set to ``"particle_layer"``.',
        type="list of .ParticleLayer",
        init_type="list of .ParticleLayer, optional",
        default="[]",
    )

    @particle_layers.validator
    def _particle_layers_validator(self, attribute, value):
        if not all(component.scale is None for component in value):
            raise ValueError(
                f"while validating {attribute.name}: components cannot be "
                "scaled individually"
            )

    _zgrid: ZGrid = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(
                lambda x: ZGrid(x) if not isinstance(x, ZGrid) else x
            ),
            validator=attrs.validators.optional(attrs.validators.instance_of(ZGrid)),
            repr=False,
        ),
        doc="A high-resolution layer altitude mesh on which the radiative "
        "properties of the components are interpolated. If unset, a default grid "
        "with one layer per 100 m (or 10 layers if the atmosphere object "
        "height is less than 100 m) is used.",
        type=".ZGrid",
        init_type=".ZGrid, optional",
    )

    @property
    def components(self) -> list[MolecularAtmosphere | ParticleLayer]:
        """
        Returns
        -------
        list of .AbstractHeterogeneousAtmosphere
            The list of all registered atmospheric components.
        """
        result = [self.molecular_atmosphere] if self.molecular_atmosphere else []
        result.extend(self.particle_layers)
        return result

    def update(self):
        # Inherit docstring
        if not self.components:
            raise ValueError("HeterogeneousAtmosphere must have at least one component")

        # Force component IDs
        for i, component in enumerate(self.components):
            component.update()
            component.id = f"{self.id}_component_{i}"
            logger.debug(
                "Override the geometry of component '%s' to %s (was %s)",
                component.id,
                self.geometry,
                component.geometry,
            )
            component.geometry = self.geometry

    # --------------------------------------------------------------------------
    #              Spatial extension and thermophysical properties
    # --------------------------------------------------------------------------

    @property
    def zgrid(self) -> ZGrid:
        # Inherit docstring
        return self._zgrid

    @property
    def bottom(self) -> pint.Quantity:
        # Inherit docstring
        return min([component.bottom for component in self.components])

    @property
    def top(self) -> pint.Quantity:
        # Inherit docstring
        return max([component.top for component in self.components])

    def spectral_set(
        self,
        quad_spec: QuadSpec | None = None,
    ) -> None | BinSet | WavelengthSet:
        if quad_spec is None:
            quad_spec = QuadSpec.default()

        components_with_spectral_set = [
            component
            for component in self.components
            if component.spectral_set(quad_spec=quad_spec) is not None
        ]

        # at most one component is allowed to have a spectral set
        if len(components_with_spectral_set) > 1:
            raise ValueError(
                "at most one component is allowed to have a spectral set",
                components_with_spectral_set,
            )

        if len(components_with_spectral_set) == 1:
            return components_with_spectral_set[0].spectral_set(quad_spec=quad_spec)

        # Fall back
        return None

    def eval_mfp(self, ctx: KernelContext) -> pint.Quantity:
        # Inherit docstring
        mfp = [component.eval_mfp(ctx=ctx) for component in self.components]
        return max(mfp)

    # --------------------------------------------------------------------------
    #                       Radiative properties
    # --------------------------------------------------------------------------

    def eval_albedo(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        if zgrid is not None and zgrid is not self.geometry.zgrid:
            raise ValueError("zgrid must be left unset or set to self.geometry.zgrid")

        units = ucc.get("collision_coefficient")
        sigma_s = self.eval_sigma_s(si).m_as(units)
        sigma_t = self.eval_sigma_t(si).m_as(units)
        albedo = np.zeros_like(sigma_s)
        np.divide(sigma_s, sigma_t, where=sigma_t != 0.0, out=albedo)

        return albedo * ureg.dimensionless

    @cache_by_id
    def _eval_sigma_t_impl(self, si: SpectralIndex) -> pint.Quantity:
        result = np.zeros((len(self.components), self.geometry.zgrid.n_layers))
        sigma_units = ucc.get("collision_coefficient")

        # Evaluate extinction for current component
        for i, component in enumerate(self.components):
            result[i] = component.eval_sigma_t(si, self.geometry.zgrid).m_as(
                sigma_units
            )

        return result * sigma_units

    def eval_sigma_t(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        if zgrid is not None and zgrid is not self.geometry.zgrid:
            raise ValueError("zgrid must be left unset or set to self.geometry.zgrid")
        return self._eval_sigma_t_impl(si).sum(axis=0)

    def eval_sigma_a(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        if zgrid is not None and zgrid is not self.geometry.zgrid:
            raise ValueError("zgrid must be left unset or set to self.geometry.zgrid")
        return self.eval_sigma_t(si) - self.eval_sigma_s(si)

    @cache_by_id
    def _eval_sigma_s_impl(self, si: SpectralIndex) -> pint.Quantity:
        result = np.zeros((len(self.components), self.geometry.zgrid.n_layers))
        sigma_units = ucc.get("collision_coefficient")

        # Evaluate scattering coefficient for current component
        for i, component in enumerate(self.components):
            result[i] = component.eval_sigma_s(si, self.geometry.zgrid).m_as(
                sigma_units
            )

        return result * sigma_units

    def _eval_sigma_s_component(
        self, si: SpectralIndex, n_component: int
    ) -> pint.Quantity:
        return self._eval_sigma_s_impl(si)[n_component]

    def eval_sigma_s(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        if zgrid is not None and zgrid is not self.geometry.zgrid:
            raise ValueError("zgrid must be left unset or set to self.geometry.zgrid")
        return self._eval_sigma_s_impl(si).sum(axis=0)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def phase(self) -> PhaseFunction:
        # Inherit docstring
        if len(self.components) == 1:
            return self.components[0].phase

        else:
            components, weights = [], []
            sigma_units = ucc.get("collision_coefficient")

            for i, component in enumerate(self.components):
                components.append(component.phase)

                def eval_sigma_s(si: SpectralIndex, n_component: int = i) -> np.ndarray:
                    return self._eval_sigma_s_component(si, n_component).m_as(
                        sigma_units
                    )

                weights.append(eval_sigma_s)

            return BlendPhaseFunction(
                components=components, weights=weights, geometry=self.geometry
            )

    @property
    def _template_phase(self) -> dict:
        # Inherit docstring
        return traverse(self.phase)[0].data

    @property
    def _params_phase(self) -> dict:
        # Inherit docstring

        umap = traverse(self.phase)[1].data

        # Add prefix and lookup strategy to all entries
        result = {}

        for uparam_key, uparam in umap.items():
            result[uparam_key] = attrs.evolve(
                uparam,
                lookup_strategy=TypeIdLookupStrategy(
                    node_type=mi.Medium,
                    node_id=self.medium_id,
                    parameter_relpath=f"phase_function.{uparam_key}",
                ),
            )

        return result



from __future__ import annotations

import attrs
import mitsuba as mi
import numpy as np

from ._core import BSDF, bsdf_factory
from ._lambertian import LambertianBSDF
from ..core import traverse
from ... import converters
from ...attrs import documented, parse_docs
from ...kernel import TypeIdLookupStrategy, UpdateParameter


def _to_bitmap(value):
    if isinstance(value, mi.Bitmap):
        return value

    elif isinstance(value, np.ndarray):
        return mi.Bitmap(value)

    elif isinstance(value, list):
        return mi.Bitmap(np.array(value))

    else:
        return value


@parse_docs
@attrs.define(eq=False, slots=False)
class OpacityMaskBSDF(BSDF):
    """
    Opacity Mask BSDF [``opacity_mask``]
    """

    opacity_bitmap: np.typing.ArrayLike | "mi.Bitmap" = documented(
        attrs.field(converter=_to_bitmap, kw_only=True),
        doc="Mitsuba bitmap that specifies the opacity of the nested BSDF "
        "plugin. This parameter has no default and is required.",
        init_type="array-like or mitsuba.Bitmap",
        type="mitsuba.Bitmap",
    )

    @opacity_bitmap.validator
    def _opacity_bitmap_validator(self, attribute, value):
        if value is not None:
            if not isinstance(value, mi.Bitmap):
                raise TypeError(
                    f"while validating '{attribute.name}': "
                    f"'{attribute.name}' must be a mitsuba Bitmap instance; "
                    f"found: {type(value)}",
                )

    uv_trafo: "mi.ScalarTransform4f" = documented(
        attrs.field(converter=converters.to_mi_scalar_transform, kw_only=True),
        doc="Transform to scale the opacity mask. This parameter has no "
        "default and is required.",
        init_type="array-like or mitsuba.ScalarTransform4f",
        type="mitsuba.ScalarTransform4f",
    )

    @uv_trafo.validator
    def _uv_trafo_validator(self, attribute, value):
        if value is not None:
            if not isinstance(value, mi.ScalarTransform4f):
                raise TypeError(
                    f"while validating '{attribute.name}': "
                    f"'{attribute.name}' must be a mitsuba ScalarTransform4f instance; "
                    f"found: {type(value)}"
                )

    nested_bsdf: BSDF = documented(
        attrs.field(
            factory=LambertianBSDF,
            converter=bsdf_factory.convert,
            validator=attrs.validators.instance_of(BSDF),
        ),
        doc="The reflection model attached to the surface.",
        type=".BSDF",
        init_type=".BSDF or dict, optional",
        default=":class:`LambertianBSDF() <.LambertianBSDF>`",
    )

    @property
    def template(self) -> dict:
        # Inherit docstring

        result = {
            "type": "mask",
            "opacity.type": "bitmap",
            "opacity.bitmap": self.opacity_bitmap,
            "opacity.filter_type": "nearest",
            "opacity.wrap_mode": "clamp",
        }

        if self.id is not None:
            result["id"] = self.id

        if self.uv_trafo is not None:
            result["opacity.to_uv"] = self.uv_trafo

        for key, value in traverse(self.nested_bsdf)[0].items():
            result[f"nested_bsdf.{key}"] = value

        return result

    @property
    def params(self) -> dict[str, UpdateParameter]:
        # Inherit Docstring

        result = {}

        # TODO: Improve this by using the BSDF's own lookup strategy
        for key, param in traverse(self.nested_bsdf)[1].items():
            result[f"nested_bsdf.{key}"] = attrs.evolve(
                param,
                lookup_strategy=TypeIdLookupStrategy(
                    node_type=mi.BSDF,
                    node_id=self.id,
                    parameter_relpath=f"nested_bsdf.{key}",
                )
                if self.id is not None
                else None,
            )

        return result



# Import accessors temporarily to set them up
from . import _accessors  # isort: skip

del _accessors

# Lazy imports
import lazy_loader  # noqa: E402

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader



from __future__ import annotations

import attrs

from ._core import PhaseFunction
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import documented, parse_docs
from ...kernel import InitParameter, UpdateParameter


@parse_docs
@attrs.define(eq=False, slots=False)
class HenyeyGreensteinPhaseFunction(PhaseFunction):
    """
    Henyey-Greenstein phase function [``hg``].

    The Henyey-Greenstein phase function :cite:`Henyey1941Diffuse` models
    scattering in an isotropic medium. The scattering pattern is controlled by
    its :math:`g` parameter, which is equal to the phase function's asymmetry
    parameter (the mean cosine of the scattering angle): a positive (resp.
    negative) value corresponds to predominant forward (resp. backward)
    scattering.
    """

    g: Spectrum = documented(
        attrs.field(
            default=0.0,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Asymmetry parameter. Must be dimensionless. "
        "Must be in :math:`]-1, 1[`.",
        type=":class:`.Spectrum`",
        init_type=":class:`.Spectrum` or dict or float, optional",
        default="0.0",
    )

    @property
    def template(self) -> dict:
        return {
            "type": "hg",
            "g": InitParameter(lambda ctx: float(self.g.eval(ctx.si))),
        }

    @property
    def params(self) -> dict[str, UpdateParameter]:
        return {
            "g": UpdateParameter(
                lambda ctx: float(self.g.eval(ctx.si)),
                UpdateParameter.Flags.SPECTRAL,
            )
        }



from ._blend import BlendPhaseFunction as BlendPhaseFunction
from ._core import PhaseFunction as PhaseFunction
from ._core import phase_function_factory as phase_function_factory
from ._hg import HenyeyGreensteinPhaseFunction as HenyeyGreensteinPhaseFunction
from ._isotropic import IsotropicPhaseFunction as IsotropicPhaseFunction
from ._rayleigh import RayleighPhaseFunction as RayleighPhaseFunction
from ._tabulated import TabulatedPhaseFunction as TabulatedPhaseFunction



from __future__ import annotations

import datetime
from abc import ABC, abstractmethod
from functools import singledispatchmethod

import attrs
import numpy as np
import pint
import pinttr
import xarray as xr
from pinttr.util import ensure_units

import eradiate

from .._factory import Factory
from ..attrs import documented, parse_docs
from ..spectral.index import CKDSpectralIndex, MonoSpectralIndex, SpectralIndex
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg

rad_profile_factory = Factory()
rad_profile_factory.register_lazy_batch(
    [
        ("_atmosphere.AtmosphereRadProfile", "atmosphere", {}),
    ],
    cls_prefix="eradiate.radprops",
)


@ureg.wraps(
    ret=None, args=("nm", "km", "km", "km^-1", "km^-1", "km^-1", ""), strict=False
)
def make_dataset(
    wavelength: pint.Quantity | float,
    z_level: pint.Quantity | float,
    z_layer: pint.Quantity | float | None = None,
    sigma_a: pint.Quantity | float | None = None,
    sigma_s: pint.Quantity | float | None = None,
    sigma_t: pint.Quantity | float | None = None,
    albedo: pint.Quantity | float | None = None,
) -> xr.Dataset:
    """
    Makes an atmospheric radiative properties data set.

    Parameters
    ----------
    wavelength : float
        Wavelength [nm].

    z_level : array
        Level altitudes [km].

    z_layer : array
        Layer altitudes [km].

        If ``None``, the layer altitudes are computed automatically, so that
        they are halfway between the adjacent altitude levels.

    sigma_a : array
        Absorption coefficient values [km^-1].

    sigma_s : array
        Scattering coefficient values [km^-1].

    sigma_t : array
        Extinction coefficient values [km^-1].

    albedo : array
        Albedo values [/].

    Returns
    -------
    Dataset
        Atmosphere radiative properties data set.
    """
    if z_layer is None:
        z_layer = 0.5 * (z_level[1:] + z_level[:-1])

    if sigma_a is not None and sigma_s is not None:
        sigma_t = sigma_a + sigma_s
        albedo = np.divide(
            sigma_s, sigma_t, where=sigma_t != 0.0, out=np.zeros_like(sigma_s)
        )
    elif sigma_t is not None and albedo is not None:
        sigma_s = albedo * sigma_t
        sigma_a = sigma_t - sigma_s
    else:
        raise ValueError(
            "You must provide either one of the two pairs of arguments "
            "('sigma_a', 'sigma_s') or ('sigma_t', 'albedo')."
        )

    return xr.Dataset(
        data_vars={
            "sigma_a": (
                ("w", "z_layer"),
                sigma_a.reshape(1, z_layer.size),
                dict(
                    standard_name="absorption_coefficient",
                    units="km^-1",
                    long_name="absorption coefficient",
                ),
            ),
            "sigma_s": (
                ("w", "z_layer"),
                sigma_s.reshape(1, z_layer.size),
                dict(
                    standard_name="scattering_coefficient",
                    units="km^-1",
                    long_name="scattering coefficient",
                ),
            ),
            "sigma_t": (
                ("w", "z_layer"),
                sigma_t.reshape(1, z_layer.size),
                dict(
                    standard_name="extinction_coefficient",
                    units="km^-1",
                    long_name="extinction coefficient",
                ),
            ),
            "albedo": (
                ("w", "z_layer"),
                albedo.reshape(1, z_layer.size),
                dict(
                    standard_name="albedo",
                    units="",
                    long_name="albedo",
                ),
            ),
        },
        coords={
            "z_level": (
                "z_level",
                z_level,
                dict(
                    standard_name="level_altitude",
                    units="km",
                    long_name="level altitude",
                ),
            ),
            "z_layer": (
                "z_layer",
                z_layer,
                dict(
                    standard_name="layer_altitude",
                    units="km",
                    long_name="layer altitude",
                ),
            ),
            "w": (
                "w",
                [wavelength],
                dict(
                    standard_name="radiation_wavelength",
                    units="nm",
                    long_name="wavelength",
                ),
            ),
        },
        attrs={
            "convention": "CF-1.8",
            "title": "Atmospheric monochromatic radiative properties",
            "history": f"{datetime.datetime.utcnow().replace(microsecond=0)} - "
            f"data set creation - "
            f"{__name__}.make_dataset",
            "source": f"eradiate, version {eradiate.__version__}",
            "references": "",
        },
    )


@parse_docs
@attrs.frozen(eq=False, init=False)
class ZGrid:
    """
    A container for a regular altitude grid.

    Notes
    -----
    * Instances are immutable.
    * Instances are hashable by ID. This is required to allow for using them as
      an argument of an LRU-cached function.
    * This class is used as the argument of the ``eval()`` family of methods.
    """

    levels: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("length"),
            on_setattr=None,  # frozen instance: on_setattr must be disabled
        ),
        type="pint.Quantity",
        init_type="quantity or array-like",
        doc="Grid node altitudes.\n"
        "\n"
        "Unit-enabled field (default: ``ucc['length']``).",
    )

    _layers: pint.Quantity = pinttr.field(
        units=ucc.deferred("length"),
        on_setattr=None,  # frozen instance: on_setattr must be disabled
    )

    _layer_height: pint.Quantity = pinttr.field(
        units=ucc.deferred("length"),
        on_setattr=None,  # frozen instance: on_setattr must be disabled
    )

    @_layer_height.validator
    def _layer_height_validator(self, attribute, value):
        if not np.isscalar(value.magnitude):
            raise ValueError("layer height must be a scalar")

    _total_height: pint.Quantity = pinttr.field(
        units=ucc.deferred("length"),
        on_setattr=None,  # frozen instance: on_setattr must be disabled
    )

    def __init__(self, levels: np.typing.ArrayLike):
        levels = ensure_units(levels, ucc.get("length"))
        layer_height = np.diff(levels)
        if not np.allclose(layer_height, layer_height[0]):
            raise ValueError("levels must be regularly spaced")
        layers = levels[:-1] + 0.5 * layer_height
        self.__attrs_init__(
            levels=levels,
            layers=layers,
            layer_height=layer_height[0],
            total_height=(levels.m[-1] - levels.m[0]) * levels.u,
        )

    @property
    def layers(self) -> pint.Quantity:
        """
        Returns
        -------
        quantity
            Vector of altitudes of layer centres.
        """
        return self._layers

    @property
    def layer_height(self) -> pint.Quantity:
        """
        Returns
        -------
        quantity
            Layer height.
        """
        return self._layer_height

    @property
    def n_levels(self) -> int:
        """
        Returns
        -------
        int
            Number of levels.
        """
        return len(self.levels)

    @property
    def n_layers(self) -> int:
        """
        Returns
        -------
        int
            Number of layers.
        """
        return len(self.layers)

    @property
    def total_height(self) -> pint.Quantity:
        """
        Returns
        -------
        quantity
            Total height covered by the altitude grid.
        """
        return self._total_height


@attrs.define(eq=False)
class RadProfile(ABC):
    """
    An abstract base class for radiative property profiles. Classes deriving
    from this one must implement methods which return the albedo and collision
    coefficients as Pint-wrapped Numpy arrays.

    See Also
    --------
    :data:`.rad_profile_factory`
    """

    @property
    @abstractmethod
    def zgrid(self) -> ZGrid:
        """
        Default altitude grid used for profile evaluation.
        """
        pass

    @singledispatchmethod
    def eval_albedo(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
    ) -> pint.Quantity:
        """
        Evaluate albedo at given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        zgrid : .ZGrid, optional
            The altitude grid for which the albedo is evaluated. If unset, a
            profile-specific default is used.

        Returns
        -------
        quantity
            Evaluated spectrum as an array with length equal to the number of
            layers.
        """
        raise NotImplementedError

    @eval_albedo.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> pint.Quantity:
        return self.eval_albedo_mono(w=si.w, zgrid=zgrid).squeeze()

    @eval_albedo.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> pint.Quantity:
        return self.eval_albedo_ckd(w=si.w, g=si.g, zgrid=zgrid).squeeze()

    def eval_albedo_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        raise NotImplementedError

    def eval_albedo_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        raise NotImplementedError

    @singledispatchmethod
    def eval_sigma_t(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
    ) -> pint.Quantity:
        """
        Evaluate extinction coefficient at given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        zgrid : .ZGrid, optional
            The altitude grid for which the extinction coefficient is evaluated.
            If unset, a profile-specific default is used.

        Returns
        -------
        quantity
            Evaluated spectrum as an array with length equal to the number of
            layers.
        """
        raise NotImplementedError

    @eval_sigma_t.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> pint.Quantity:
        return self.eval_sigma_t_mono(w=si.w, zgrid=zgrid).squeeze()

    @eval_sigma_t.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> pint.Quantity:
        return self.eval_sigma_t_ckd(w=si.w, g=si.g, zgrid=zgrid).squeeze()

    def eval_sigma_t_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        raise NotImplementedError

    def eval_sigma_t_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        raise NotImplementedError

    @singledispatchmethod
    def eval_sigma_a(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
    ) -> pint.Quantity:
        """
        Evaluate absorption coefficient at given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        zgrid : .ZGrid, optional
            The altitude grid for which the absorption coefficient is evaluated.
            If unset, a profile-specific default is used.

        Returns
        -------
        quantity
            Evaluated spectrum as an array with length equal to the number of
            layers.
        """
        raise NotImplementedError

    @eval_sigma_a.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> pint.Quantity:
        return self.eval_sigma_a_mono(w=si.w, zgrid=zgrid).squeeze()

    @eval_sigma_a.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> pint.Quantity:
        return self.eval_sigma_a_ckd(w=si.w, g=si.g, zgrid=zgrid).squeeze()

    def eval_sigma_a_mono(
        self,
        w: pint.Quantity,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        """
        Evaluate absorption coefficient spectrum in monochromatic mode.
        """
        raise NotImplementedError

    def eval_sigma_a_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        """
        Evaluate absorption coefficient spectrum in CKD modes.
        """
        raise NotImplementedError

    @singledispatchmethod
    def eval_sigma_s(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
    ) -> pint.Quantity:
        """
        Evaluate scattering coefficient at given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        zgrid : :class:`.ZGrid`, optional
            The altitude grid for which the scattering coefficient is evaluated.
            If unset, a profile-specific default is used.

        Returns
        -------
        quantity
            Evaluated spectrum as an array with length equal to the number of
            layers.
        """
        raise NotImplementedError

    @eval_sigma_s.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> pint.Quantity:
        return self.eval_sigma_s_mono(w=si.w, zgrid=zgrid).squeeze()

    @eval_sigma_s.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> pint.Quantity:
        return self.eval_sigma_s_ckd(w=si.w, g=si.g, zgrid=zgrid).squeeze()

    def eval_sigma_s_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        raise NotImplementedError

    def eval_sigma_s_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        raise NotImplementedError

    @singledispatchmethod
    def eval_dataset(
        self,
        si: SpectralIndex,
        zgrid: ZGrid | None = None,
    ) -> xr.Dataset:
        """
        Evaluate radiative properties at given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        zgrid : .ZGrid, optional
            The altitude grid for which the radiative profile is evaluated.
            If unset, a profile-specific default is used.

        Returns
        -------
        Dataset
            Radiative property dataset.
        """
        raise NotImplementedError

    @eval_dataset.register(MonoSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> xr.Dataset:
        return self.eval_dataset_mono(w=si.w, zgrid=zgrid)

    @eval_dataset.register(CKDSpectralIndex)
    def _(self, si, zgrid: ZGrid) -> xr.Dataset:
        return self.eval_dataset_ckd(w=si.w, g=si.g, zgrid=zgrid)

    def eval_dataset_mono(self, w: pint.Quantity, zgrid: ZGrid) -> xr.Dataset:
        raise NotImplementedError

    def eval_dataset_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> xr.Dataset:
        raise NotImplementedError



import attrs

from ._core import PhaseFunction
from ...attrs import parse_docs


@parse_docs
@attrs.define(eq=False, slots=False)
class IsotropicPhaseFunction(PhaseFunction):
    """
    Isotropic phase function [``isotropic``].

    The isotropic phase function models scattering with equal probability in
    all directions.
    """

    @property
    def template(self) -> dict:
        return {"type": "isotropic"}



"""
Mathematical and physical constants.
"""

from .units import unit_registry as ureg

#: Earth radius
EARTH_RADIUS = 6378.1 * ureg.km

#: Lower bound of the default spectral range
SPECTRAL_RANGE_MIN = 280.0 * ureg.nm

#: Upper bound of the default spectral range
SPECTRAL_RANGE_MAX = 2400.0 * ureg.nm



# This is a vendored and updated copy of the asvdb project (https://github.com/rapidsai/asvdb).

import glob
import json
import os
import random
import stat
import tempfile
import time
from os import path
from pathlib import Path
from urllib.parse import urlparse

BenchmarkInfoKeys = set(
    [
        "machineName",
        "cudaVer",
        "osType",
        "pythonVer",
        "commitHash",
        "commitTime",
        "branch",
        "gpuType",
        "cpuType",
        "num_cpu",
        "arch",
        "ram",
        "gpuRam",
        "requirements",
        "env_vars",
    ]
)

BenchmarkResultKeys = set(
    [
        "funcName",
        "result",
        "argNameValuePairs",
        "unit",
    ]
)


class BenchmarkInfo:
    """
    Meta-data describing the environment for a benchmark or set of benchmarks.
    """

    def __init__(
        self,
        machineName="",
        cudaVer="",
        osType="",
        pythonVer="",
        commitHash="",
        commitTime=0,
        branch="",
        envName="",
        gpuType="",
        cpuType="",
        numCpu="",
        arch="",
        ram="",
        gpuRam="",
        requirements=None,
        envVar=None,
        resultColumns=None,
    ):
        self.machineName = machineName
        self.cudaVer = cudaVer
        self.osType = osType
        self.gpuType = gpuType
        self.cpuType = cpuType
        self.numCpu = numCpu
        self.arch = arch
        self.ram = ram
        self.gpuRam = gpuRam

        self.pythonVer = pythonVer
        self.commitHash = commitHash
        self.commitTime = int(commitTime)
        self.branch = branch
        self.envName = envName

        self.requirements = requirements or {}
        self.envVar = envVar or {}

        self.resultColumns = resultColumns or []

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(machineName='{self.machineName}'\n"
            f", cudaVer='{self.cudaVer}'\n"
            f", osType='{self.osType}'\n"
            f", pythonVer='{self.pythonVer}'\n"
            f", commitHash='{self.commitHash}'\n"
            f", commitTime={self.commitTime}\n"
            f", branch={self.branch}\n"
            f", envName={self.envName}\n"
            f", gpuType='{self.gpuType}'\n"
            f", cpuType='{self.cpuType}'\n"
            f", numCpu='{self.numCpu}'\n"
            f", arch='{self.arch}'\n"
            f", ram={repr(self.ram)}\n"
            f", gpuRam={repr(self.gpuRam)}\n"
            f", requirements={repr(self.requirements)}\n"
            f", envVar={repr(self.envVar)}\n"
            f", resultColumns={repr(self.resultColumns)}\n"
            ")"
        )

    def __eq__(self, other):
        return (
            (self.machineName == other.machineName)
            and (self.cudaVer == other.cudaVer)
            and (self.osType == other.osType)
            and (self.pythonVer == other.pythonVer)
            and (self.commitHash == other.commitHash)
            and (self.commitTime == other.commitTime)
            and (self.branch == other.branch)
            and (self.envName == other.envName)
            and (self.gpuType == other.gpuType)
            and (self.cpuType == other.cpuType)
            and (self.numCpu == other.numCpu)
            and (self.arch == other.arch)
            and (self.ram == other.ram)
            and (self.gpuRam == other.gpuRam)
            and (self.requirements == other.requirements)
            and (self.envVar == other.envVar)
            and (self.resultColumns == other.resultColumns)
        )


class BenchmarkResult:
    """
    The result of a benchmark run for a particular benchmark function, given
    specific args.
    """

    def __init__(
        self,
        funcName,
        results,
        paramNames=None,
        param=None,
        argNameValuePairs=None,
        unit=None,
        benchType=None,
        code=None,
        version=None,
        minRunCount=2,
        number=0,
        repeat=0,
        rounds=2,
        sampleTime=0.01,
        warmupTime=-1,
    ):
        self.funcName = funcName
        self.argNameValuePairs = self.__sanitizeArgNameValues(argNameValuePairs)

        self.benchType = benchType or "time"
        self.unit = unit or "seconds"
        self.code = code or funcName

        if type(results) is not list:
            results = [results]

        self.paramNames = paramNames or []
        self.param = param or []
        self.version = version or "2"

        self.minRunCount = minRunCount
        self.number = number
        self.repeat = repeat
        self.rounds = rounds
        self.sampleTime = sampleTime
        self.warmupTime = warmupTime

        self.results = results

    def __sanitizeArgNameValues(self, argNameValuePairs):
        if argNameValuePairs is None:
            return []
        return [(n, str(v if v is not None else "NaN")) for (n, v) in argNameValuePairs]

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(funcName='{self.funcName}'\n"
            f", results={repr(self.results)}\n"
            f", argNameValuePairs={repr(self.argNameValuePairs)}\n"
            f", unit='{self.unit}'\n"
            f", type='{self.benchType}'\n"
            f", param='{self.param}'\n"
            f", paramNames='{self.paramNames}'\n"
            f", version='{self.version}'\n"
            f", minRunCount='{self.minRunCount}'\n"
            f", number='{self.number}'\n"
            f", repeat='{self.repeat}'\n"
            f", rounds='{self.rounds}'\n"
            f", sampleTime='{self.sampleTime}'\n"
            f", warmupTime='{self.warmupTime}'\n"
            f", results='{self.results}'\n"
            ")"
        )

    def __eq__(self, other):
        return (
            (self.funcName == other.funcName)
            and (self.argNameValuePairs == other.argNameValuePairs)
            and (self.results == other.result)
            and (self.unit == other.unit)
            and (self.code == other.code)
            and (self.param == other.param)
            and (self.paramNames == other.paramNames)
            and (self.version == other.version)
            and (self.minRunCount == other.minRunCount)
            and (self.number == other.number)
            and (self.repeat == other.repeat)
            and (self.rounds == other.rounds)
            and (self.sampleTime == other.sampleTime)
            and (self.warmupTime == other.warmupTime)
            and (self.results == other.results)
        )


class ASVDb:
    """
    A "database" of benchmark results consumable by ASV.
    https://asv.readthedocs.io/en/stable/dev.html?highlight=%24results_dir#benchmark-suite-layout-and-file-formats
    """

    confFileName = "asv.conf.json"
    defaultResultsDirName = "results"
    defaultHtmlDirName = "html"
    defaultConfVersion = 1
    benchmarksFileName = "benchmarks.json"
    machineFileName = "machine.json"
    lockfilePrefix = ".asvdbLOCK"

    def __init__(
        self, dbDir, repo=None, branches=None, projectName=None, commitUrl=None
    ):
        """
        dbDir - directory containing the ASV results, config file, etc.
        repo - the repo associated with all reasults in the DB.
        branches - https://asv.readthedocs.io/en/stable/asv.conf.json.html#branches
        projectName - the name of the project to display in ASV reports
        commitUrl - the URL ASV will use in reports to redirect users to when
                    they click on a data point. This is typically a Github
                    project URL that shows the contents of a commit.
        """
        self.dbDir = dbDir
        self.repo = repo
        self.branches = branches
        self.projectName = projectName
        self.commitUrl = commitUrl

        self.machineFileExt = path.join(
            self.defaultResultsDirName, "*", self.machineFileName
        )
        self.confFileExt = self.confFileName
        self.confFilePath = path.join(self.dbDir, self.confFileName)
        self.confVersion = self.defaultConfVersion
        self.resultsDirName = self.defaultResultsDirName
        self.resultsDirPath = path.join(self.dbDir, self.resultsDirName)
        self.htmlDirName = self.defaultHtmlDirName
        self.benchmarksFileExt = path.join(
            self.defaultResultsDirName, self.benchmarksFileName
        )
        self.benchmarksFilePath = path.join(
            self.resultsDirPath, self.benchmarksFileName
        )

        # Each ASVDb instance must have a unique lockfile name to identify other
        # instances that may be setting locks.
        self.lockfileName = "%s-%s-%s" % (self.lockfilePrefix, os.getpid(), time.time())
        self.lockfileTimeout = 5  # seconds

        # S3-related attributes
        if self.__isS3URL(dbDir):
            self.s3Resource = None  # boto3.resource("s3")
            self.bucketName = urlparse(self.dbDir, allow_fragments=False).netloc
            self.bucketKey = urlparse(self.dbDir, allow_fragments=False).path.lstrip(
                "/"
            )

        ########################################
        # Testing and debug members
        self.debugPrint = False
        # adds a delay during write operations to easily test write collision
        # handling.
        self.writeDelay = 0
        # To "cancel" write operations that are being delayed.
        self.cancelWrite = False

    ###########################################################################
    # Public API
    ###########################################################################
    def loadConfFile(self):
        """
        Read the ASV conf file on disk and set - or possibly overwrite - the
        member variables with the contents of the file.
        """
        self.__assertDbDirExists()
        try:
            self.__getLock(self.dbDir)
            # FIXME: check if confFile exists
            self.__downloadIfS3()

            d = self.__loadJsonDictFromFile(self.confFilePath)
            self.resultsDirName = d.get("results_dir", self.resultsDirName)
            self.resultsDirPath = path.join(self.dbDir, self.resultsDirName)
            self.benchmarksFilePath = path.join(
                self.resultsDirPath, self.benchmarksFileName
            )
            self.htmlDirName = d.get("html_dir", self.htmlDirName)
            self.repo = d.get("repo")
            self.branches = d.get("branches", [])
            self.projectName = d.get("project")
            self.commitUrl = d.get("show_commit_url")

            self.__uploadIfS3()

        finally:
            self.__releaseLock(self.dbDir)
            self.__removeLocalS3Copy()

    def updateConfFile(self):
        """
        Update the ASV conf file with the values passed in to the CTOR. This
        also ensures the object is up-to-date with any changes to the conf file
        that may have been done by other ASVDb instances.
        """
        self.__ensureDbDirExists()
        try:
            self.__getLock(self.dbDir)
            self.__downloadIfS3()

            if self.__waitForWrite():
                self.__updateConfFile()

            self.__uploadIfS3()

        finally:
            self.__releaseLock(self.dbDir)
            self.__removeLocalS3Copy()

    def addResult(self, benchmarkInfo, benchmarkResult):
        """
        Add the benchmarkResult associated with the benchmarkInfo to the DB.
        This will also update the conf file with the CTOR args if not done
        already.
        """
        self.__ensureDbDirExists()
        try:
            self.__getLock(self.dbDir)
            self.__downloadIfS3(bInfo=benchmarkInfo)
            if self.__waitForWrite():
                self.__updateFilesForInfo(benchmarkInfo)
                self.__updateFilesForResult(benchmarkInfo, benchmarkResult)

            self.__uploadIfS3()

        finally:
            self.__releaseLock(self.dbDir)
            self.__removeLocalS3Copy()

    def addResults(self, benchmarkInfo, benchmarkResultList):
        """
        Add each benchmarkResult obj in benchmarkResultList associated with
        benchmarkInfo to the DB.  This will also update the conf file with the
        CTOR args if not done already.
        """
        self.__ensureDbDirExists()
        try:
            self.__getLock(self.dbDir)
            self.__downloadIfS3(bInfo=benchmarkInfo)

            if self.__waitForWrite():
                self.__updateFilesForInfo(benchmarkInfo)
                for resultObj in benchmarkResultList:
                    self.__updateFilesForResult(benchmarkInfo, resultObj)

            self.__uploadIfS3()

        finally:
            self.__releaseLock(self.dbDir)
            self.__removeLocalS3Copy()

    def getInfo(self):
        """
        Return a list of BenchmarkInfo objs from reading the db files on disk.
        """
        self.__assertDbDirExists()
        try:
            self.__getLock(self.dbDir)
            self.__downloadIfS3()
            retList = self.__readResults(infoOnly=True)

        finally:
            self.__releaseLock(self.dbDir)
            self.__removeLocalS3Copy()

        return retList

    def getResults(self, filterInfoObjList=None):
        """
        Return a list of (BenchmarkInfo obj, [BenchmarkResult obj, ...]) tuples
        from reading the db files on disk.  filterInfoObjList is expected to be
        a list of BenchmarkInfo objs, and if provided will be used to return
        results for only those BenchmarkInfo objs.
        """
        self.__assertDbDirExists()
        try:
            self.__getLock(self.dbDir)
            self.__downloadIfS3(results=True)
            retList = self.__readResults(filterByInfoObjs=filterInfoObjList)

        finally:
            self.__releaseLock(self.dbDir)
            self.__removeLocalS3Copy()

        return retList

    ###########################################################################
    # Private methods. These should not be called by clients. Among other
    # things, public methods use proper locking to ensure atomic operations
    # and these do not.
    ###########################################################################
    def __readResults(self, infoOnly=False, filterByInfoObjs=None):
        """
        Main "read" method responsible for reading ASV JSON files and creating
        BenchmarkInfo and BenchmarkResult objs.

        If infoOnly==True, returns a list of only BenchmarkInfo objs, otherwise
        returns a list of tuples containing (BenchmarkInfo obj, [BenchmarkResult
        obj, ...]) to represent each BenchmarkInfo object and all the
        BenchmarkResult objs associated with it.

        filterByInfoObjs can be set to only return BenchmarkInfo objs and their
        results that match at least one of the BenchmarkInfo objs in the
        filterByInfoObjs list (the list is treated as ORd).
        """
        retList = []

        resultsPath = Path(self.resultsDirPath)

        # benchmarks.json containes meta-data about the individual benchmarks,
        # which is only needed for returning results.
        if not (infoOnly):
            benchmarksJsonFile = resultsPath / self.benchmarksFileName
            if benchmarksJsonFile.exists():
                bDict = self.__loadJsonDictFromFile(benchmarksJsonFile.as_posix())
            else:
                # FIXME: test
                raise FileNotFoundError(f"{benchmarksJsonFile.as_posix()}")

        for machineDir in resultsPath.iterdir():
            # Each subdir under the results dir contains all results for a
            # individual machine. The only non-dir (file) that may need to be
            # read in the results dir is benchmarks.json, which would have been
            # read above.
            if machineDir.is_dir():
                # Inside the individual machine dir, ;ook for and read
                # machine.json first.  Assume this is not a valid results dir if
                # no machine file and skip.
                machineJsonFile = machineDir / self.machineFileName
                if machineJsonFile.exists():
                    mDict = self.__loadJsonDictFromFile(machineJsonFile.as_posix())
                else:
                    continue

                # Read each results file and populate the machineResults list.
                # This will contain either BenchmarkInfo objs or tuples of
                # (BenchmarkInfo, [BenchmarkResult objs, ...]) based on infoOnly
                machineResults = []
                for resultsFile in machineDir.iterdir():
                    if resultsFile == machineJsonFile:
                        continue
                    rDict = self.__loadJsonDictFromFile(resultsFile.as_posix())

                    resultsParams = rDict.get("params", {})
                    # Each results file has a single BenchmarkInfo obj
                    # describing it.
                    bi = BenchmarkInfo(
                        machineName=mDict.get("machine", ""),
                        cudaVer=resultsParams.get("cuda", ""),
                        osType=resultsParams.get("os", ""),
                        pythonVer=resultsParams.get("python", ""),
                        commitHash=rDict.get("commit_hash", ""),
                        commitTime=rDict.get("date", ""),
                        branch=rDict.get("branch", ""),
                        envName=rDict.get("env_name", ""),
                        gpuType=mDict.get("gpu", ""),
                        cpuType=mDict.get("cpu", ""),
                        numCpu=mDict.get("num_cpu", ""),
                        arch=mDict.get("arch", ""),
                        ram=mDict.get("ram", ""),
                        gpuRam=mDict.get("gpuRam", ""),
                        requirements=rDict.get("requirements", {}),
                        envVar=rDict.get("env_vars", {}),
                        resultColumns=rDict.get("result_columns", None),
                    )

                    # If a filter was specified, at least one EXACT MATCH to the
                    # BenchmarkInfo obj must be present.
                    if filterByInfoObjs and bi not in filterByInfoObjs:
                        continue

                    if infoOnly:
                        machineResults.append(bi)
                    else:
                        # FIXME: if results not in rDict, throw better error
                        resultsDict = rDict["results"]
                        # Populate the list of BenchmarkResult objs associated
                        # with the BenchmarkInfo obj
                        resultObjs = []
                        for benchmarkName in resultsDict:
                            # benchmarkSpec is the entry in benchmarks.json,
                            # which is needed for the param names
                            if benchmarkName not in bDict:
                                print(
                                    "WARNING: Encountered benchmark name "
                                    "that is not in "
                                    f"{self.benchmarksFileName}: "
                                    f"file: {resultsFile.as_posix()} "
                                    f'invalid name"{benchmarkName}", skipping.'
                                )
                                continue

                            # the number of result columns should always be the same as the number of results
                            # not true for because samples and profile only exist if the benchmark is run with profile mode
                            # assert len(bi.resultColumns) == len(resultsDict[benchmarkName])

                            benchmarkSpec = bDict[benchmarkName]
                            # benchmarkResults is the entry in this particular
                            # result file for this benchmark
                            # benchmarkResults = dict(zip(rDict["result_columns"],resultsDict[benchmarkName]))

                            # dropping the parameter picking. we will just overwrite
                            br = BenchmarkResult(
                                funcName=benchmarkName,
                                results=resultsDict[benchmarkName],
                                paramNames=benchmarkSpec["param_names"],
                                param=benchmarkSpec["params"],
                                code=benchmarkSpec["code"],
                                version=benchmarkSpec["version"],
                                benchType=benchmarkSpec["type"],
                                unit=benchmarkSpec["unit"],
                                minRunCount=benchmarkSpec["min_run_count"],
                                number=benchmarkSpec["number"],
                                repeat=benchmarkSpec["repeat"],
                                rounds=benchmarkSpec["rounds"],
                                sampleTime=benchmarkSpec["sample_time"],
                                warmupTime=benchmarkSpec["warmup_time"],
                            )
                            unit = benchmarkSpec.get("unit")
                            code = benchmarkSpec.get("code")
                            if unit is not None:
                                br.unit = unit
                            if code is not None:
                                br.code = code
                            resultObjs.append(br)

                            # NM 18/07/2024: Disabled parameter expansion
                            # -------------------------------------------

                            # Inverse of the write operation described in
                            # self.__updateResultJson()
                            # paramsCartProd = list(itertools.product(*paramValues))
                            # for (paramValueCombo, result) in zip(paramsCartProd, results):
                            # br = BenchmarkResult(
                            # funcName=benchmarkName,
                            # argNameValuePairs=zip(paramNames, paramValueCombo),
                            # result=result)
                            #     unit = benchmarkSpec.get("unit")
                            #     if unit is not None:
                            #         br.unit = unit
                            #     resultObjs.append(br)
                        machineResults.append((bi, resultObjs))

                retList += machineResults

        return retList

    def __updateFilesForInfo(self, benchmarkInfo):
        """
        Updates all the db files that are affected by a new BenchmarkInfo obj.
        """
        # special case: if the benchmarkInfo has a new branch specified,
        # update self.branches so the conf files includes the new branch
        # name.
        newBranch = benchmarkInfo.branch
        if newBranch and newBranch not in self.branches:
            self.branches.append(newBranch)

        # The comments below assume default dirname values (mainly
        # "results"), which can be changed in the asv.conf.json file.
        #
        # <self.dbDir>/asv.conf.json
        self.__updateConfFile()
        # <self.dbDir>/results/<machine dir>/machine.json
        self.__updateMachineJson(benchmarkInfo)

    def __updateFilesForResult(self, benchmarkInfo, benchmarkResult):
        """
        Updates all the db files that are affected by a new BenchmarkResult
        obj. This also requires the corresponding BenchmarkInfo obj since some
        results files also include info data.
        """
        # <self.dbDir>/results/benchmarks.json
        self.__updateBenchmarkJson(benchmarkResult)
        # <self.dbDir>/results/<machine dir>/<result file name>.json
        self.__updateResultJson(benchmarkResult, benchmarkInfo)

    def __assertDbDirExists(self):
        # FIXME: update for support S3 - this method should return True if
        # self.dbDir is a valid S3 URL or a valid path on disk.
        if self.__isS3URL(self.dbDir):
            self.s3Resource.Bucket(self.bucketName).objects
        else:
            if not (path.isdir(self.dbDir)):
                raise FileNotFoundError(
                    f"{self.dbDir} does not exist or is " "not a directory"
                )

    def __ensureDbDirExists(self):
        # FIXME: for S3 support, if self.dbDir is a S3 URL then simply check if
        # it's valid and exists, but don't try to create it (raise an exception
        # if it does not exist).  For a local file path, create it if it does
        # not exist, like already being done below.
        if self.__isS3URL(self.dbDir):
            self.s3Resource.Bucket(self.bucketName).objects
        else:
            if not (path.exists(self.dbDir)):
                os.mkdir(self.dbDir)
                # Hack: os.mkdir() seems to return before the filesystem catches up,
                # so pause before returning to help ensure the dir actually exists
                time.sleep(0.1)

    def __updateConfFile(self):
        """
        Update the conf file with the settings in this ASVDb instance.
        """
        if self.repo is None:
            raise AttributeError(
                "repo must be set to non-None before " f"writing {self.confFilePath}"
            )

        d = self.__loadJsonDictFromFile(self.confFilePath)
        # ASVDb is git-only for now, so ensure .git extension
        d["repo"] = self.repo + (".git" if not self.repo.endswith(".git") else "")
        currentBranches = d.get("branches", [])
        d["branches"] = currentBranches + [
            b for b in (self.branches or []) if b not in currentBranches
        ]
        d["version"] = self.confVersion
        d["project"] = self.projectName or self.repo.replace(".git", "").split("/")[-1]
        d["show_commit_url"] = self.commitUrl or (
            self.repo.replace(".git", "")
            + ("/" if not self.repo.endswith("/") else "")
            + "commit/"
        )

        self.__writeJsonDictToFile(d, self.confFilePath)

    def __updateBenchmarkJson(self, benchmarkResult):
        # The following is an example of the schema ASV expects for
        # `benchmarks.json`.  If param names are A, B, and C
        #
        # {
        #     "<algo name>": {
        #         "code": "",
        #         "name": "<algo name>",
        #         "param_names": [
        #             "A", "B", "C"
        #         ],
        #     "params": [
        #                [<value1 for A>,
        #                 <value2 for A>,
        #                ],
        #                [<value1 for B>,
        #                 <value2 for B>,
        #                ],
        #                [<value1 for C>,
        #                 <value2 for C>,
        #                ],
        #               ],
        #         "timeout": 60,
        #         "type": "time",
        #         "unit": "seconds",
        #         "version": 1,
        #     }
        # }

        newParamNames = []
        newParamValues = []
        for n, v in benchmarkResult.argNameValuePairs:
            newParamNames.append(n)
            newParamValues.append(v)

        d = self.__loadJsonDictFromFile(self.benchmarksFilePath)

        benchDict = d.setdefault(
            benchmarkResult.funcName,
            self.__getDefaultBenchmarkDescrDict(
                benchmarkResult.funcName, newParamNames
            ),
        )
        benchDict["unit"] = benchmarkResult.unit
        benchDict["type"] = benchmarkResult.benchType
        benchDict["code"] = benchmarkResult.code
        benchDict["version"] = benchmarkResult.version

        benchDict["param_names"] = benchmarkResult.paramNames
        benchDict["params"] = benchmarkResult.param

        benchDict["min_run_count"] = benchmarkResult.minRunCount
        benchDict["number"] = benchmarkResult.number
        benchDict["repeat"] = benchmarkResult.repeat
        benchDict["rounds"] = benchmarkResult.rounds
        benchDict["sample_time"] = benchmarkResult.sampleTime
        benchDict["warmup_time"] = benchmarkResult.warmupTime

        # NM 18/07/2024: Disabled parameter expansion
        # -------------------------------------------

        # existingParamNames = benchDict["param_names"]
        # existingParamValues = benchDict["params"]

        # numExistingParams = len(existingParamNames)
        # numExistingParamValues = len(existingParamValues)
        # numNewParams = len(newParamNames)

        # # Check for the case where a result came in for the function, but it has
        # # a different number of args vs. what was saved previously
        # if numExistingParams != numNewParams:
        #     raise ValueError("result for %s had %d params in benchmarks.json, "
        #                      "but new result has %d params" \
        #                      % (benchmarkResult.funcName, numExistingParams,
        #                         numNewParams))
        # numParams = numNewParams

        # cartProd = list(itertools.product(*existingParamValues))
        # if tuple(newParamValues) not in cartProd:
        #     if numExistingParamValues == 0:
        #         for newVal in newParamValues:
        #             existingParamValues.append([newVal])
        #     else:
        #         for i in range(numParams):
        #             if newParamValues[i] not in existingParamValues[i]:
        #                 existingParamValues[i].append(newParamValues[i])

        d[benchmarkResult.funcName] = benchDict

        # a version key must always be present in self.benchmarksFilePath,
        # "current" ASV version requires this to be 2 (or higher?)
        # d["version"] = benchmarkResult["version"]
        d["version"] = 2
        self.__writeJsonDictToFile(d, self.benchmarksFilePath)

    def __updateMachineJson(self, benchmarkInfo):
        # The following is an example of the schema ASV expects for
        # `machine.json`.
        # {
        #     "arch": "x86_64",
        #     "cpu": "Intel, ...",
        #     "machine": "sm01",
        #     "os": "Linux ...",
        #     "ram": "123456",
        #     "version": 1,
        # }

        machineFilePath = path.join(
            self.resultsDirPath, benchmarkInfo.machineName, self.machineFileName
        )
        d = self.__loadJsonDictFromFile(machineFilePath)
        d["arch"] = benchmarkInfo.arch
        d["cpu"] = benchmarkInfo.cpuType
        d["gpu"] = benchmarkInfo.gpuType
        # d["cuda"] = benchmarkInfo.cudaVer
        d["machine"] = benchmarkInfo.machineName
        # d["os"] = benchmarkInfo.osType
        d["ram"] = benchmarkInfo.ram
        d["gpuRam"] = benchmarkInfo.gpuRam
        d["version"] = 1
        self.__writeJsonDictToFile(d, machineFilePath)

    def __updateResultJson(self, benchmarkResult, benchmarkInfo):
        # The following is an example of the schema ASV expects for
        # '<machine>-<commit_hash>.json'. If param names are A, B, and C
        #
        # {
        #     "params": {
        #         "cuda": "9.2",
        #         "gpu": "Tesla ...",
        #         "machine": "sm01",
        #         "os": "Linux ...",
        #         "python": "3.7",
        #     },
        #     "requirements": {},
        #     "result_columns":["result", "params", "version", "started_at", "duration", "stats_ci_99_a", "stats_ci_99_b", "stats_q_25", "stats_q_75", "stats_number", "stats_repeat", "samples", "profile"]
        #     "results": {
        #         "<algo name>": {
        #              [[<result1>,<result2>,],[<param1>, <param2>],"<version>",...]
        #     },
        #     "commit_hash": "321e321321eaf",
        #     "date": 12345678,
        #     "python": "3.7",
        #     "version": 1,
        # }

        resultsFilePath = self.__getResultsFilePath(benchmarkInfo)
        d = self.__loadJsonDictFromFile(resultsFilePath)

        d["commit_hash"] = benchmarkInfo.commitHash
        d["env_name"] = benchmarkInfo.envName
        d["date"] = int(benchmarkInfo.commitTime)
        d["python"] = benchmarkInfo.pythonVer
        d["params"] = {
            "arch": benchmarkInfo.arch,
            "cpu": benchmarkInfo.cpuType,
            "num_cpu": benchmarkInfo.numCpu,
            "machine": benchmarkInfo.machineName,
            "os": benchmarkInfo.osType,
            "python": benchmarkInfo.pythonVer,
        }
        d["requirements"] = benchmarkInfo.requirements
        d["env_vars"] = benchmarkInfo.envVar
        d["result_columns"] = benchmarkInfo.resultColumns

        allResultsDict = d.setdefault("results", {})
        allResultsDict[benchmarkResult.funcName] = benchmarkResult.results
        d["version"] = 2

        # NM 18/07/2024: Disabled parameter expansion
        # -------------------------------------------

        # resultDict = allResultsDict.setdefault(benchmarkResult.funcName, {})

        # existingParamValuesList = benchmarkInfo.params
        # existingResultValueList = resultDict.setdefault("result", [])

        # ASV uses the cartesian product of the param values for looking up the
        # result for a particular combination of param values.  For example:
        # "params": [["a"], ["b", "c"], ["d", "e"]] results in: [("a", "b",
        # "d"), ("a", "b", "e"), ("a", "c", "d"), ("a", "c", "e")] and each
        # combination of param values has a result, with the results for the
        # corresponding param values in the same order.  If a result for a set
        # of param values DNE, use None.

        # store existing results in map based on cartesian product of all
        # current params.
        # paramsCartProd = list(itertools.product(*existingParamValuesList))
        # Assume there is an equal number of results for cartProd values
        # (some will be None)
        # paramsResultMap = dict(zip(paramsCartProd, existingResultValueList))

        # FIXME: dont assume these are ordered properly (ie. the same way as
        # defined in benchmarks.json)
        # newResultParamValues = tuple(v for (_, v) in benchmarkResult.argNameValuePairs)

        # Update the "params" lists with the new param settings for the new result.
        # Only add values that are not already present
        # numExistingParamValues = len(existingParamValuesList)
        # if numExistingParamValues == 0:
        #     for newParamValue in newResultParamValues:
        #         existingParamValuesList.append([newParamValue])
        #     results = [benchmarkResult.results]

        # else:
        #     for i in range(numExistingParamValues):
        #         if newResultParamValues[i] not in existingParamValuesList[i]:
        #             existingParamValuesList[i].append(newResultParamValues[i])

        #     # Add the new result
        #     paramsResultMap[newResultParamValues] = benchmarkResult.results

        #     # Re-compute the cartesian product of all param values now that the
        #     # new values are added. Use this to determine where to place the new
        #     # result in the result list.
        #     results = []
        #     for paramVals in itertools.product(*existingParamValuesList):
        #         results.append(paramsResultMap.get(paramVals))

        # resultDict["params"] = existingParamValuesList
        # resultDict["results"] = benchmarkResult.results

        self.__writeJsonDictToFile(d, resultsFilePath)

    def __getDefaultBenchmarkDescrDict(self, funcName, paramNames):
        return {
            "code": funcName,
            "min_run_count": 2,
            "name": funcName,
            "number": 0,
            "param_names": paramNames,
            "params": [],
            "repeat": 0,
            "rounds": 1,
            "sample_time": 0.01,
            "type": "time",
            "unit": "seconds",
            "version": 2,
            "warmup_time": -1,
        }

    def __getResultsFilePath(self, benchmarkInfo):
        # The path to the resultsFile will be based on additional params present
        # in the benchmarkInfo obj.
        assert (
            len(benchmarkInfo.commitHash[:8]) >= 8
        ), "commit hash needs to be at leas 8 characters long"

        fileNameParts = [
            benchmarkInfo.commitHash[:8],
            benchmarkInfo.envName,
        ]
        fileName = "-".join(fileNameParts) + ".json"
        return path.join(self.resultsDirPath, benchmarkInfo.machineName, fileName)

    def __loadJsonDictFromFile(self, jsonFile):
        """
        Return a dictionary representing the contents of jsonFile by
        either reading in the existing file or returning {}
        """
        if path.exists(jsonFile):
            with open(jsonFile) as fobj:
                # FIXME: ideally this could use flock(), but some situations do
                # not allow grabbing a file lock (NFS?)
                # fcntl.flock(fobj, fcntl.LOCK_EX)
                # FIXME: error checking
                return json.load(fobj)

        return {}

    def __writeJsonDictToFile(self, jsonDict, filePath):
        # FIXME: error checking
        dirPath = path.dirname(filePath)
        if not path.isdir(dirPath):
            os.makedirs(dirPath)

        with open(filePath, "w") as fobj:
            # FIXME: ideally this could use flock(), but some situations do not
            # allow grabbing a file lock (NFS?)
            # fcntl.flock(fobj, fcntl.LOCK_EX)
            json.dump(jsonDict, fobj, indent=2)

    ###########################################################################
    # ASVDb private locking methods
    ###########################################################################
    def __getLock(self, dirPath):
        if self.__isS3URL(dirPath):
            self.__getS3Lock()
        else:
            self.__getLocalFileLock(dirPath)

    def __getLocalFileLock(self, dirPath):
        """
        Gets a lock on dirPath against other ASVDb instances (in other
        processes, possibily on other machines) using the following technique:

        * Check for other locks and clear them if they've been seen for longer
          than self.lockfileTimeout (do this to help cleanup after others that
          may have died prematurely)

        * Once all locks are clear - either by their owner because they
          finished their read/write, or by removing them because they're
          presumed dead - create a lock for this instance

        * If a race condition was detected, probably because multiple ASVDbs
          saw all locks were cleared at the same time and created their locks
          at the same time, remove this lock, and wait a random amount of time
          before trying again.  The random time prevents yet another race.
        """
        otherLockfileTimes = {}
        thisLockfile = path.join(dirPath, self.lockfileName)
        # FIXME: This shouldn't be needed? But if so, be smarter about
        # preventing an infintite loop?
        i = 0
        while i < 1000:
            # Keep checking for other locks to clear
            self.__updateOtherLockfileTimes(dirPath, otherLockfileTimes)
            # FIXME: potential infintite loop due to starvation?
            otherLocks = list(otherLockfileTimes.keys())
            while otherLocks:
                if self.debugPrint:
                    print(
                        f"This lock file will be {thisLockfile} but other "
                        f"locks present: {otherLocks}, waiting to try to "
                        "lock again..."
                    )
                time.sleep(0.2)
                self.__updateOtherLockfileTimes(dirPath, otherLockfileTimes)
                otherLocks = list(otherLockfileTimes.keys())

            # All clear, create lock
            if self.debugPrint:
                print(f"All clear, setting lock {thisLockfile}")
            self.__createLockfile(dirPath)

            # Check for a race condition where another lock could have been created
            # while creating the lock for this instance.
            self.__updateOtherLockfileTimes(dirPath, otherLockfileTimes)

            # If another lock snuck in while this instance was creating its
            # lock, remove this lock and wait a random amount of time before
            # trying again (random time to prevent another race condition with
            # the competing instance, this way someone will clearly get there
            # first)
            if otherLockfileTimes:
                self.__releaseLock(dirPath)
                randTime = (int(5 * random.random()) + 1) + random.random()
                if self.debugPrint:
                    print(
                        f"Collision - waiting {randTime} seconds before "
                        "trying to lock again."
                    )
                time.sleep(randTime)
            else:
                break

            i += 1

    def __releaseLock(self, dirPath):
        if self.__isS3URL(dirPath):
            self.__releaseS3Lock()
        else:
            self.__releaseLocalFileLock(dirPath)

    def __releaseLocalFileLock(self, dirPath):
        thisLockfile = path.join(dirPath, self.lockfileName)
        if self.debugPrint:
            print(f"Removing lock {thisLockfile}")
        self.__removeFiles([thisLockfile])

    def __updateOtherLockfileTimes(self, dirPath, lockfileTimes):
        """
        Return a list of lockfiles that have "timed out", probably because their
        process was killed. This will never include the lockfile for this
        instance.  Update the lockfileTimes dict as a side effect with the
        discovery time of any new lockfiles and remove any lockfiles that are no
        longer present.
        """
        thisLockfile = path.join(dirPath, self.lockfileName)
        now = time.time()
        expired = []

        allLockfiles = glob.glob(path.join(dirPath, self.lockfilePrefix) + "*")

        if self.debugPrint:
            print(
                f"   This lockfile is {thisLockfile}, allLockfiles is "
                f"{allLockfiles}, lockfileTimes is {lockfileTimes}"
            )
        # Remove lockfiles from the lockfileTimes dict that are no longer
        # present on disk
        lockfilesToRemove = set(lockfileTimes.keys()) - set(allLockfiles)
        for removedLockfile in lockfilesToRemove:
            lockfileTimes.pop(removedLockfile)

        # check for expired lockfiles while also setting the discovery time on
        # new lockfiles in the lockfileTimes dict.
        for lockfile in allLockfiles:
            if lockfile == thisLockfile:
                continue
            if (now - lockfileTimes.setdefault(lockfile, now)) > self.lockfileTimeout:
                expired.append(lockfile)

        if self.debugPrint:
            print(
                f"   This lockfile is {thisLockfile}, lockfileTimes is "
                f"{lockfileTimes}, now is {now}, expired is {expired}"
            )
        self.__removeFiles(expired)

    def __createLockfile(self, dirPath):
        """
        low-level lockfile creation - consider calling __getLock() instead.
        """
        thisLockfile = path.join(dirPath, self.lockfileName)
        open(thisLockfile, "w").close()
        # Make the lockfile read/write to all so others can remove it if this
        # process dies prematurely
        os.chmod(
            thisLockfile,
            (
                stat.S_IRUSR
                | stat.S_IWUSR
                | stat.S_IRGRP
                | stat.S_IWGRP
                | stat.S_IROTH
                | stat.S_IWOTH
            ),
        )

    ###########################################################################
    # S3 Locking methods
    ###########################################################################
    def __getS3Lock(self):
        thisLockfile = path.join(self.bucketKey, self.lockfileName)
        # FIXME: This shouldn't be needed? But if so, be smarter about
        # preventing an infintite loop?
        i = 0

        # otherLockfileTimes is a tuple representing (<List of lockfiles>, <Length of List>)
        otherLockfileTimes = ([], 0)
        while i < 1000:
            otherLockfileTimes = self.__updateS3LockfileTimes()
            debugCounter = 0

            while otherLockfileTimes[1] != 0:
                if self.debugPrint:
                    lockfileList = []
                    for each in otherLockfileTimes[0]:
                        lockfileList.append(each.key)

                    print(
                        f"This lock file will be {thisLockfile} but other "
                        f"locks present: {lockfileList}, waiting to try to "
                        "lock again..."
                    )

                time.sleep(1)
                otherLockfileTimes = self.__updateS3LockfileTimes()

            # All clear, create lock
            if self.debugPrint:
                print(f"All clear, setting lock {thisLockfile}")
            self.s3Resource.Object(self.bucketName, thisLockfile).put()

            # Give S3 time to see the new lock
            time.sleep(1)

            # Check for a race condition where another lock could have been created
            # while creating the lock for this instance.
            otherLockfileTimes = self.__updateS3LockfileTimes()

            if otherLockfileTimes[1] != 0:
                self.__releaseS3Lock()
                randTime = (int(30 * random.random()) + 5) + random.random()
                if self.debugPrint:
                    print(
                        f"Collision - waiting {randTime} seconds before "
                        "trying to lock again."
                    )
                time.sleep(randTime)
            else:
                break

            i += 1

    def __updateS3LockfileTimes(self):
        # Find lockfiles in S3 Bucket
        response = self.s3Resource.Bucket(self.bucketName).objects.filter(
            Prefix=path.join(self.bucketKey, self.lockfilePrefix)
        )

        length = 0
        for lockfile in response:
            length += 1
            if self.lockfileName in lockfile.key:
                lockfile.delete()
                length -= 1

        return (response, length)

    def __releaseS3Lock(self):
        thisLockfile = path.join(self.bucketKey, self.lockfileName)
        if self.debugPrint:
            print(f"Removing lock {thisLockfile}")
        self.s3Resource.Object(self.bucketName, thisLockfile).delete()

    ###########################################################################
    # S3 utilities
    ###########################################################################
    def __downloadIfS3(self, bInfo=BenchmarkInfo(), results=False):
        def downloadS3(bucket, ext):
            bucket.download_file(
                path.join(self.bucketKey, ext), path.join(self.localS3Copy.name, ext)
            )

        if not self.__isS3URL(self.dbDir):
            return

        self.localS3Copy = tempfile.TemporaryDirectory()
        os.makedirs(path.join(self.localS3Copy.name, self.defaultResultsDirName))
        bucket = self.s3Resource.Bucket(self.bucketName)
        # If results isn't set, only download key files, else download key files and results
        if results == False:
            keyFileExts = [
                self.confFileExt,
                self.machineFileExt,
                self.benchmarksFileExt,
            ]
            # Use Try/Except to catch file Not Found errors and continue, avoids additional API calls
            for fileExt in keyFileExts:
                try:
                    downloadS3(bucket, fileExt)
                except Exception as e:
                    err = "Not Found"
                    if err not in e.response["Error"]["Message"]:
                        raise

            # Download specific result file for updating results if BenchmarkInfo is sent
            try:
                if bInfo.machineName != "":
                    commitHash, pyVer, cuVer, osType = (
                        bInfo.commitHash,
                        bInfo.pythonVer,
                        bInfo.cudaVer,
                        bInfo.osType,
                    )
                    filename = f"{commitHash}-python{pyVer}-cuda{cuVer}-{osType}.json"
                    os.makedirs(
                        path.join(
                            self.localS3Copy.name,
                            self.defaultResultsDirName,
                            bInfo.machineName,
                        ),
                        exist_ok=True,
                    )
                    resultFileExt = path.join(
                        self.defaultResultsDirName, bInfo.machineName, filename
                    )
                    downloadS3(bucket, resultFileExt)

            except Exception as e:
                err = "Not Found"
                if err not in e.response["Error"]["Message"]:
                    raise

        else:
            try:
                downloadS3(bucket, self.confFileExt)
            except Exception as e:
                err = "Not Found"
                if err not in e.response["Error"]["Message"]:
                    raise

            try:
                resultsBucketPath = path.join(
                    self.bucketKey, self.defaultResultsDirName
                )
                resultsLocalPath = path.join(
                    self.localS3Copy.name, self.defaultResultsDirName
                )

                # Loop over ASV results folder and download everything.
                # objectExt represents the file extension starting from the base resultsBucketPath
                # For example: resultsBucketPath = "asvdb/results"
                #            : objectKey = "asvdb/results/machine_name/results.json
                #            : objectExt = "machine_name/results.json"
                for bucketObj in bucket.objects.filter(Prefix=resultsBucketPath):
                    objectExt = bucketObj.key.replace(resultsBucketPath + "/", "")
                    if len(objectExt.split("/")) > 1:
                        os.makedirs(
                            path.join(resultsLocalPath, objectExt.split("/")[0]),
                            exist_ok=True,
                        )
                    bucket.download_file(
                        bucketObj.key, path.join(resultsLocalPath, objectExt)
                    )

            except Exception as e:
                err = "Not Found"
                if err not in e.response["Error"]["Message"]:
                    raise e

        # Set all the internal locations to point to the downloaded files:
        self.confFilePath = path.join(self.localS3Copy.name, self.confFileName)
        self.resultsDirPath = path.join(self.localS3Copy.name, self.resultsDirName)
        self.benchmarksFilePath = path.join(
            self.resultsDirPath, self.benchmarksFileName
        )

    def __uploadIfS3(self):
        def recursiveUpload(base, ext=""):
            root, dirs, files = next(os.walk(path.join(base, ext), topdown=True))

            # Upload files in this folder
            for name in files:
                self.s3Resource.Bucket(self.bucketName).upload_file(
                    path.join(base, ext, name), path.join(self.bucketKey, ext, name)
                )

            # Call upload again for each folder
            if len(dirs) != 0:
                for folder in dirs:
                    ext = path.join(ext, folder)
                    recursiveUpload(base, ext)

        if self.__isS3URL(self.dbDir):
            recursiveUpload(self.localS3Copy.name)
            # Give S3 time to see the new uploads before releasing lock
            time.sleep(1)

    def __removeLocalS3Copy(self):
        if not self.__isS3URL(self.dbDir):
            return

        self.localS3Copy.cleanup()
        self.localS3Copy = None

        self.confFilePath = path.join(self.dbDir, self.confFileName)
        self.resultsDirPath = path.join(self.dbDir, self.resultsDirName)
        self.benchmarksFilePath = path.join(
            self.resultsDirPath, self.benchmarksFileName
        )

    ###########################################################################
    def __removeFiles(self, fileList):
        for f in fileList:
            try:
                os.remove(f)
            except FileNotFoundError:
                pass

    def __isS3URL(self, url):
        """
        Returns True if url is a S3 URL, False otherwise.
        """
        if url.startswith("s3:"):
            return True

        return False

    def __waitForWrite(self):
        """
        Testing helper: pause for self.writeDelay seconds, or until
        self.cancelWrite turns True. Always set self.cancelWrite back to False
        so future writes can take place by default.

        Return True to indicate a write operation should take place, False to
        cancel the write operation, based on if the write was cancelled or not.
        """
        if not (self.cancelWrite):
            st = now = time.time()
            while ((now - st) < self.writeDelay) and not (self.cancelWrite):
                time.sleep(0.01)
                now = time.time()

        retVal = not (self.cancelWrite)
        self.cancelWrite = False
        return retVal



"""
Geometric transforms.
"""
from __future__ import annotations

import mitsuba as mi
import numpy as np


def map_unit_cube(
    xmin: float, xmax: float, ymin: float, ymax: float, zmin: float, zmax: float
) -> mi.ScalarTransform4f:
    r"""
    Map the unit cube :math:`[0, 1]^3` to
    :math:`[x_\mathrm{min}, x_\mathrm{max}] \times [y_\mathrm{min}, y_\mathrm{max}] \times [z_\mathrm{min}, z_\mathrm{max}]`.

    Parameters
    ----------
    xmin : float
        Minimum X value.

    xmax : float
        Maximum X value.

    ymin : float
        Minimum Y value.

    ymax : float
        Maximum Y value.

    zmin : float
        Minimum Z value.

    zmax : float
        Maximum Z value.

    Returns
    -------
    :class:`mitsuba.core.ScalarTransform4f`
        Computed transform matrix.

    Warnings
    --------
    You must select a Mitsuba variant before calling this function.
    """
    return mi.ScalarTransform4f.translate(
        [xmin, ymin, zmin]
    ) @ mi.ScalarTransform4f.scale([xmax - xmin, ymax - ymin, zmax - zmin])


def map_cube(
    xmin: float, xmax: float, ymin: float, ymax: float, zmin: float, zmax: float
) -> mi.ScalarTransform4f:
    r"""
    Map the cube :math:`[-1, 1]^3` to
    :math:`[x_\mathrm{min}, x_\mathrm{max}] \times [y_\mathrm{min}, y_\mathrm{max}] \times [z_\mathrm{min}, z_\mathrm{max}]`.

    Parameters
    ----------
    xmin : float
        Minimum X value.

    xmax : float
        Maximum X value.

    ymin : float
        Minimum Y value.

    ymax : float
        Maximum Y value.

    zmin : float
        Minimum Z value.

    zmax : float
        Maximum Z value.

    Returns
    -------
    :class:`mitsuba.core.ScalarTransform4f`
        Computed transform matrix.

    Warnings
    --------
    You must select a Mitsuba variant before calling this function.
    """
    half_edge_x = 0.5 * (xmax - xmin)
    half_edge_y = 0.5 * (ymax - ymin)
    half_edge_z = 0.5 * (zmax - zmin)

    return mi.ScalarTransform4f.translate(
        [half_edge_x + xmin, half_edge_y + ymin, half_edge_z + zmin]
    ) @ mi.ScalarTransform4f.scale([half_edge_x, half_edge_y, half_edge_z])


def transform_affine(t, x):
    """
    Apply a mitsuba transformation to a list of points.

    Parameters
    ----------
    t : mi.ScalarTransform4f
        A transformation, to be applied to the list of points.

    x : array-like
        A list of points in (n, 3)-shape, to be transformed.

    Returns
    -------
    ndarray
        List in (n, 3)-shape of transformed points.
    """
    t = np.array(t.matrix)

    r = np.dot(t, (np.vstack((x.T, np.ones(x.shape[0])))))
    r /= r[3, :]
    r = r[:3, ...].T

    return r
